# coding: utf-8

from time import time
from re import findall
from typing import Union, List, Optional
from .inout import InOut
from ..tool import ft232h_server as ft232h, Usb_server
from .._global import OptionalModule

try:
  from smbus2 import SMBus
except (ModuleNotFoundError, ImportError):
  smbus2 = OptionalModule("smbus2")

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  import board
except (ModuleNotFoundError, ImportError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  import busio
except (ModuleNotFoundError, ImportError):
  busio = OptionalModule('busio', 'Blinka is necessary to use the I2C bus')

try:
  import adafruit_ads1x15.ads1115 as ads
  from adafruit_ads1x15.analog_in import AnalogIn
except (ModuleNotFoundError, ImportError):
  ADS = OptionalModule('adafruit_ads1x15',
                       'Blinka is necessary to use the I2C bus')
  AnalogIn = OptionalModule('AnalogIn',
                            'Blinka is necessary to use the I2C bus')

# Register and other configuration values:
Ads1115_pointer_conversion = 0x00
Ads1115_pointer_config = 0x01
Ads1115_pointer_lo_thresh = 0x02
Ads1115_pointer_hi_thresh = 0x03

Ads1115_config_mux = {'A0 - A1': 0x0000,
                      'A0 - A3': 0x1000,
                      'A1 - A3': 0x2000,
                      'A2 - A3': 0x3000,
                      'A0': 0x4000,
                      'A1': 0x5000,
                      'A2': 0x6000,
                      'A3': 0x7000}

Ads1115_config_gain = {0.256: 0x0a00,
                       0.512: 0x0800,
                       1.024: 0x0600,
                       2.048: 0x0400,
                       4.096: 0x0200,
                       6.144: 0x0000}

Ads1115_config_dr = {8: 0x0000,
                     16: 0x0020,
                     32: 0x0040,
                     64: 0x0060,
                     128: 0x0080,
                     250: 0x00A0,
                     475: 0x00C0,
                     860: 0x00E0}

Ads1115_blinka_gain = {0.256: 16,
                       0.512: 8,
                       1.024: 4,
                       2.048: 2,
                       4.096: 1,
                       6.144: 2/3}

Ads1115_backends = ['Pi4', 'ft232h', 'blinka']


class Ads1115(Usb_server, InOut):
  """A class for controlling Adafruit's ADS1115 16-bits ADC.

  The Ads1115 InOut block is meant for reading output values from a 16-bits
  ADS1115 ADC, using the I2C protocol. The output is in Volts by default, but a
  ``gain`` and an ``offset`` can be specified.
  """

  def __init__(self,
               backend: str,
               device_address: int = 0x48,
               i2c_port: int = 1,
               sample_rate: int = 128,
               v_range: float = 2.048,
               multiplexer: str = 'A1',
               dry_pin: Optional[Union[str, int]] = None,
               gain: float = 1,
               offset: float = 0,
               ft232h_ser_num: Optional[str] = None) -> None:
    """Checks arguments validity.

    Args:
      backend (:obj:`str`): The backend for communicating with the ADS1115.
        Should be one of:
        ::

          'Pi4', 'ft232h', 'blinka'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing Adafruit's
        modules, but these modules are compatible with and maintained on a wide
        variety of boards. The `'ft232h'` backend allows controlling the
        ADS1115 from a PC using Adafruit's FT232H USB to I2C adapter. See
        :ref:`Crappy for embedded hardware` for details.
      device_address (:obj:`int`, optional): The I2C address of the ADS1115.
        The default address is `0x48`, but it is possible to change this
        setting using the `ADDR` pin.
      i2c_port (:obj:`int`, optional): The I2C port over which the ADS1115
        should communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      sample_rate (:obj:`int`, optional): The sample rate for data conversion
        (in SPS). Available sample rates are:
        ::

          8, 16, 32, 64, 128, 250, 475, 860

      v_range (:obj:`float`, optional): The value (in Volts) of the measured
        signal corresponding to the `0x7FFF` output in bits, i.e. that
        saturates the sensor. A signal of ``-v_range`` Volts gives a `0x8000`
        output in bits. Available ``v_range`` values are:
        ::

          0.256, 0.512, 1.024, 2.048, 4.096, 6.144

      multiplexer (:obj:`str`, optional): Choice of the inputs to consider.
        Single-input modes actually measure `Ax - GND`. The available
        ``multiplexer`` values are:
        ::

          'A0', 'A1', 'A2', 'A3',
          'A0 - A1',
          'A0 - A3',
          'A1 - A3',
          'A2 - A3'

      dry_pin (:obj:`int` or :obj:`str`, optional): Optionally, reads the end
        of conversion signal from a GPIO rather than from an I2C message.
        Speeds up the reading and decreases the traffic on the bus, but
        requires one extra wire. With the backend `'Pi4'`, give the index of
        the GPIO in BCM convention. With the `'ft232h'` backend, give the name
        of the GPIO in the format `Dx` or `Cx`. This feature is not available
        with the `'blinka'` backend.
      gain (:obj:`float`, optional): Allows to tune the output value according
        to the formula:
        ::

          output = gain * tension + offset.

      offset (:obj:`float`, optional): Allows to tune the output value
        according to the formula:
        ::

          output = gain * tension + offset.

      ft232h_ser_num (:obj:`str`, optional): If backend is `'ft232h'`, the
        serial number of the FT232H to use for communication.

    Warning:
      AINx voltages should not be higher than `VDD+0.3V` nor lower than
      `GND-0.3V`. Setting high ``v_range`` values does not allow measuring
      voltages higher than `VDD` !!
    """

    if not isinstance(backend, str) or backend not in Ads1115_backends:
      raise ValueError("backend should be in {}".format(Ads1115_backends))
    self._backend = backend

    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    InOut.__init__(self)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    if backend == 'Pi4':
      self._bus = SMBus(i2c_port)
    elif backend == 'ft232h':
      self._bus = ft232h(mode='I2C',
                         block_number=block_number,
                         current_file=current_file,
                         command_file=command_file,
                         answer_file=answer_file,
                         block_lock=block_lock,
                         current_lock=current_lock,
                         serial_nr=ft232h_ser_num)
    elif backend == 'blinka':
      i2c = busio.I2C(board.SCL, board.SDA)
      self._ads = ads.ADS1115(i2c)

    self._device_address = device_address

    if v_range not in Ads1115_config_gain:
      raise ValueError("v_range should be in {}".format(list(
        Ads1115_config_gain.keys())))
    else:
      self._v_range = v_range

    if sample_rate not in Ads1115_config_dr:
      raise ValueError("sample_rate should be in {}".format(list(
        Ads1115_config_dr.keys())))
    else:
      self._sample_rate = sample_rate

    if multiplexer not in Ads1115_config_mux:
      raise ValueError("multiplexer should be in {}".format(list(
        Ads1115_config_mux.keys())))
    else:
      self._multiplexer = multiplexer

    if dry_pin is not None:
      if backend == 'ft232h' and not isinstance(dry_pin, str):
        raise TypeError('int_pin should be a string when using the ft232h '
                        'backend !')
      elif backend == 'Pi4' and not isinstance(dry_pin, int):
        raise TypeError('int_pin should be an int when using the Pi4 '
                        'backend !')
      elif backend == 'blinka' and dry_pin is not None:
        print("Warning ! Getting the end of conversion information from the"
              "DRY pin is not supported by the blinka backend.")
    self._dry_pin = dry_pin

    self._gain = gain
    self._offset = offset

  def open(self) -> None:
    """Sets the I2C communication and device."""

    if self._backend == 'blinka':
      self._ads.gain = Ads1115_blinka_gain[self._v_range]
      self._ads.data_rate = self._sample_rate
      chan_nr = findall(r'\d', self._multiplexer)
      if len(chan_nr) == 1:
        self._chan = AnalogIn(self._ads,
                              getattr(ads, 'P{}'.format(chan_nr[0])))
      else:
        self._chan = AnalogIn(self._ads,
                              getattr(ads, 'P{}'.format(chan_nr[0])),
                              getattr(ads, 'P{}'.format(chan_nr[1])))

    else:
      if not self._is_connected():
        raise IOError("The ADS1115 is not connected")

      # Setting the configuration register according to the user values
      init_value = Ads1115_config_mux[self._multiplexer]
      init_value |= Ads1115_config_gain[self._v_range]
      init_value |= 0x0100  # Single shot operating mode
      init_value |= Ads1115_config_dr[self._sample_rate]
      if self._dry_pin is None:
        # Setting the two last bits to 11 to disable the DRY pin
        init_value |= 0x0003

      self._set_register(Ads1115_pointer_config, init_value)

      # Setting the threshold registers to activate the DRY pin output
      if self._dry_pin is not None:
        self._set_register(Ads1115_pointer_lo_thresh, 0x0000)
        self._set_register(Ads1115_pointer_hi_thresh, 0xFFFF)

      if self._backend == 'Pi4' and self._dry_pin is not None:
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._dry_pin, GPIO.IN)

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in Volts, unless a gain and offset are applied.

    Returns:
      :obj:`list`: A list containing the timeframe and the voltage value
    """

    if self._backend == 'blinka':
      out = [time()]
      value = self._chan.voltage

    else:
      # Reading the config register, and setting it so that the ADS1115 starts
      # a conversion
      ms_byte, ls_byte = self._bus.read_i2c_block_data(self._device_address,
                                                       Ads1115_pointer_config,
                                                       2)
      self._set_register(Ads1115_pointer_config,
                         ((ms_byte | 0x80) << 8) | ls_byte)

      # Waiting for the end of the conversion
      t0 = time()
      while not self._data_available():
        if time() - t0 > 0.5:
          raise TimeoutError('Waited too long for data to be ready !')

      # Reading the output of the conversion
      out = [time()]
      ms_byte, ls_byte = self._bus.read_i2c_block_data(
          self._device_address,
          Ads1115_pointer_conversion,
          2)

      # Converting the output value into Volts
      value_raw = (ms_byte << 8) | ls_byte
      if ms_byte >> 7:
        value_raw -= 2 ** 16
      value = value_raw * self._v_range / 2 ** 15

    out.append(self._offset + self._gain * value)
    return out

  def close(self) -> None:
    """Closes the I2C bus"""

    if self._backend != 'blinka':
      self._bus.close()

    if self._backend == 'Pi4' and self._dry_pin is not None:
      GPIO.cleanup()

  def _set_register(self, register_address: int, value: int) -> None:
    """Thin wrapper for writing data to the registers."""

    self._bus.write_i2c_block_data(self._device_address, register_address,
                                   [(value >> 8) & 0xFF, value & 0xFF])

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns:
      :obj:`bool`: :obj:`True` if reading was successful, else :obj:`False`
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._dry_pin is None:
      return self._bus.read_i2c_block_data(self._device_address,
                                           Ads1115_pointer_config,
                                           1)[0] & 0x80
    # EOC signal from a GPIO
    elif self._backend == 'ft232h':
      return not bool(self._bus.get_gpio(self._dry_pin))
    else:
      return not bool(GPIO.input(self._dry_pin))



# coding: utf-8

from time import time, sleep
from typing import Optional, Dict, Any, Union, List
import numpy as np

from .._global import DefinitionError


class MetaIO(type):
  """Metaclass ensuring that two InOuts don't have the same name, and that all
  InOuts define the required methods. Also keeps track of all the InOut
  classes, including the custom user-defined ones."""

  classes = {}

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that an InOut with the same name doesn't already exist
    if name in cls.classes:
      raise DefinitionError(f"The {name} class is already defined !")

    # Saving the class
    if name != 'InOut':
      cls.classes[name] = cls


class InOut(metaclass=MetaIO):
  """Base class for all InOut objects. Implements methods shared by all the
  these objects, and ensures their dataclass is MetaIO."""

  def __init__(self) -> None:
    """Sets the attributes."""

    self._compensations = list()

  def open(self) -> None:
    """This method should perform any action that's required for initializing
    the hardware and the communication with it.

    Communication with hardware should be avoided in the :meth:`__init__`
    method, and this method is where it should start happening. This method is
    called after Crappy's processes start, i.e. when the associated IOBlock
    already runs separately from all the other blocks.
    """

    ...

  def get_data(self) -> Optional[Union[list, Dict[str, Any]]]:
    """This method should acquire data from a device and return it in a
    :obj:`list` along with a timestamp.

    The timestamp must always be the first returned value, and there can be any
    number of other acquired channels. The same number of values should always
    be returned, and they should be in the same order.

    Alternatively, the values can be returned in a :obj:`dict`. In that case,
    the ``labels`` argument of the IOBlock is ignored and the returned labels
    correspond to the keys of the dict.

    It is alright for this method to return :obj:`None` if there's no data to
    acquire.
    """

    print(f"WARNING ! The get_data method is not defined for the InOut "
          f"{type(self).__name__} !\n To get rid of this warning, define a "
          f"get_data method, or remove all the downstream links and make sure "
          f"the make_zero_delay argument of the IOBlock is set to None.")
    sleep(1)
    return

  def set_cmd(self, *cmd) -> None:
    """This method should handle commands received from the upstream blocks.

    Usually the command is meant to be set on a device, but any other behavior
    is possible. The commands will be passed to this method as `args` (not
    `kwargs`), in the same order as the ``cmd_labels`` are given in the
    IOBlock.

    If the expected number of commands is always the same, you can simply put
    as many `args` to your ``set_cmd`` method as there are commands. For
    example for three commands:
    ::

      def set_cmd(self, cmd0, cmd1, cmd2):
        ...

    Alternatively, or if the number of commands may vary from one test to
    another, you can get all the commands at once in a :obj:`tuple` by putting
    a single unpacking argument. Example:
    ::

      def set_cmd(self, *cmds):
        number_of_commands = len(cmds)
        cmd0 = cmds[0]
        ...

    """

    print(f"WARNING ! The InOut {type(self).__name__} has incoming links but"
          f" its set_cmd method is not defined !\nThe data received from the "
          f"incoming links is discarded.")
    sleep(1)
    return

  def start_stream(self) -> None:
    """This method should start the acquisition of the stream."""

    print(f"WARNING ! The InOut {type(self).__name__} does not define the "
          f"start_stream method !")

  def get_stream(self) -> Optional[List[np.ndarray]]:
    """This method should acquire a stream as a numpy array, and return it in a
    :obj:`list` along with an array carrying the timestamps.

    The time array must be the first element of the list, the stream array the
    second element. The time array should have only one column, the stream
    array can have any number of columns representing the different channels
    acquired.

    It is also possible to return the two arrays in a :obj:`dict`, in which
    case the ``labels`` argument is ignored and the keys of the dict set the
    returned labels.

    It is alright for this method to return :obj:`None` if there's no data to
    acquire.
    """

    print(f"WARNING ! The InOut {type(self).__name__} has downstream links but"
          f" its get_stream method is not defined !\nNo data sent to "
          f"downstream links.")
    sleep(1)
    return

  def stop_stream(self) -> None:
    """This method should stop the acquisition of the stream."""

    print(f"WARNING ! The InOut {type(self).__name__} does not define the "
          f"stop_stream method !")

  def close(self) -> None:
    """This method should perform any action required for properly ending the
    test and closing the communication with hardware.

    It will be called when the associated IOBlock receives the order to stop,
    either because the user hit CTRL+C, or because a Generator block reached
    the end of its path, or because an exception was raised in any of the
    blocks.
    """

    ...

  def make_zero(self, delay: float) -> None:
    """Acquires data for a given delay, averages it for each channel, and
    stores the average.

    Does not work for pure streams, as it requires a :meth:`get_data` for
    acquiring the data.

    The average values will then be used to remove the offset of the acquired
    data during the test.
    """

    buf = []
    t0 = time()

    # Acquiring data for a given delay
    while time() < t0 + delay:
      data = self.get_data()
      if data is not None and len(data) > 1:
        buf.append(data[1:])

    # If no data could be acquired, abort
    if not buf:
      print(f"No data acquired when zeroing the channels of the InOut "
            f"{type(self).__name__}, aborting the zeroing.")
      return

    # Averaging the values and storing them
    for values in zip(*buf):
      try:
        self._compensations.append(-sum(values) / len(values))
      except TypeError:
        # If something goes wrong, just forget about the offsetting
        self._compensations = list()
        print(f"WARNING ! Cannot calculate the offset for the InOut "
              f"{type(self).__name__} !\nPossible reasons are that it doesn't"
              f"return only numbers, or that it returns a dict instead of the "
              f"expected list.")
        return

  def return_data(self) -> Optional[Union[list, Dict[str, Any]]]:
    """Returns the data from :meth:`get_data`, corrected by an offset if the
    ``make_zero_delay`` argument of the IOBlock is set."""

    data = self.get_data()

    # If there's no offsetting, just return the data
    if data is None or not self._compensations:
      return data

    # Otherwise, offset the acquired data except for time
    elif len(data[1:]) == len(self._compensations):
      try:
        return [data[0]] + [val + comp for val, comp in
                            zip(data[1:], self._compensations)]
      # Shouldn't happen but doesn't harm to be careful
      except TypeError:
        return data

    # Should also not happen
    else:
      raise ValueError("The number of offsets doesn't match the number of "
                       "acquired values.")

  def return_stream(self) -> Optional[List[np.ndarray]]:
    """Returns the data from :meth:`get_stream`, corrected by an offset if the
    ``make_zero_delay`` argument of the IOBlock is set."""

    data = self.get_stream()

    # If there's no offsetting, just return the data
    if data is None or not self._compensations:
      return data

    # Otherwise, offset the acquired data except for time
    elif data[1].shape[1] == len(self._compensations):
      try:
        return [data[0], data[1] + self._compensations]
      # Shouldn't happen but doesn't harm to be careful
      except TypeError:
        return data

    # There's a problem with the shape of the output data
    else:
      raise ValueError("The number of offsets doesn't match the shape of the "
                       "acquired array.")



# coding: utf-8

"""
Example showing how to use thermocouples with an NI daq board.

Required hardware:
  - Nidaqmx compatible board
  - Thermocouple(s)
"""

import crappy

STREAMER = True
channels = range(1)
chan_names = ['cDAQ1Mod1/ai%d' % i for i in channels]
labels = ['t(s)'] + ['T%d' % i for i in channels]

if __name__ == "__main__":
  io = crappy.blocks.IOBlock("Nidaqmx",
                             channels=[dict(name=c_name, units='C',
                                            type='thrmcpl',
                                            thermocouple_type='K')
                                       for c_name in chan_names],
                             samplerate=14. / len(channels),
                             labels=['t(s)', 'stream'] if STREAMER else labels,
                             streamer=STREAMER)

  graph = crappy.blocks.Grapher(*[('t(s)', lab) for lab in labels[1:]])
  if STREAMER:
    crappy.link(io, graph,
                modifier=crappy.modifier.Demux(labels[1:],
                                               mean=False, transpose=True))
  else:
    crappy.link(io, graph)
  crappy.start()



# coding: utf-8

from time import time, sleep
from typing import Tuple, Any
from .camera import Camera
from .._global import OptionalModule
import numpy as np
from threading import Thread, RLock

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")

try:
  from picamera import PiCamera
  from picamera.array import PiRGBArray
except (ModuleNotFoundError, ImportError, OSError):
  PiCamera = OptionalModule("picamera")

picamera_iso = [0, 100, 200, 320, 400, 500, 640, 800]


class Picamera(Camera):
  """Class for reading images from a PiCamera.

  The Picamera Camera block is meant for reading images from a Picamera.
  It uses the :mod:`picamera` module for capturing images, and :mod:`cv2` for
  converting bgr images to black and white.

  Warning:
    Only works on Raspberry Pi, with the picamera API. On the latest OS release
    "Bullseye", it has to be specifically activated in the configuration menu.
  """

  def __init__(self) -> None:
    """Instantiates the available settings."""

    super().__init__()
    self._cam = PiCamera()

    # Settings definition
    self.add_scale_setting('Width', 1, 3280, self._get_width,
                           self._set_width, 1280)
    self.add_scale_setting('Height', 1, 2464, self._get_height,
                           self._set_height, 720)
    self.add_scale_setting('Iso (0 for auto)', 0, 800, self._get_iso,
                           self._set_iso, 0)
    self.add_scale_setting('Brightness', 0, 100, self._get_brightness,
                           self._set_brightness, 50)
    self.add_scale_setting('Contrast', -100, 100, self._get_contrast,
                           self._set_contrast, 0)
    self.add_scale_setting('Saturation', -100, 100, self._get_saturation,
                           self._set_saturation, 0)
    self.add_scale_setting('Shutter speed (0 for auto)', 0, 30,
                           self._get_shutter_speed, self._set_shutter_speed, 0)
    self.add_bool_setting('Black_and_white', self._get_black_white,
                          self._set_black_white, True)
    self.add_scale_setting('Crop: X offset', 0.0, 1.0, self._get_crop_x_offset,
                           self._set_crop_x_offset, 0.0)
    self.add_scale_setting('Crop: Y offset', 0.0, 1.0, self._get_crop_y_offset,
                           self._set_crop_y_offset, 0.0)
    self.add_scale_setting('Crop: width', 0.0, 1.0, self._get_crop_width,
                           self._set_crop_width, 1.0)
    self.add_scale_setting('Crop: height', 0.0, 1.0, self._get_crop_height,
                           self._set_crop_height, 1.0)

    self._frame_grabber = Thread(target=self._grab_frame)
    self._lock = RLock()
    self._frame = None
    self._stop = False
    self._started = False

  def open(self, **kwargs: Any) -> None:
    """Sets the settings to their default values and starts the thread."""

    self.set_all(**kwargs)

    # Starting the video stream
    self._capture = PiRGBArray(self._cam, (self._get_width(),
                                           self._get_height()))
    self._stream = self._cam.capture_continuous(self._capture, format='bgr',
                                                use_video_port=True)

    self._frame_grabber.start()
    sleep(1)
    self._started = True

  def _stop_stream(self) -> None:
    """Stops the video stream before changing the image size."""

    if not self._started:
      return

    self._stop = True
    self._frame_grabber.join()
    self._capture.close()

  def _restart_stream(self) -> None:
    """Restarts the video stream after a change in the image size."""

    if not self._started:
      return

    self._stop = False
    self._frame_grabber = Thread(target=self._grab_frame)
    self._capture = PiRGBArray(self._cam, (self._get_width(),
                                           self._get_height()))
    self._stream = self._cam.capture_continuous(self._capture, format='bgr',
                                                use_video_port=True)

    self._frame_grabber.start()
    sleep(1)

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Simply returns the last image in the buffer.

    The captured image is in bgr format, and converted into black and white if
    needed.

    Returns:
      The timeframe and the image.
    """

    t = time()
    with self._lock:
      output = self._frame
    if self.Black_and_white:
      output = cv2.cvtColor(output, cv2.COLOR_BGR2GRAY)
    return t, output

  def _grab_frame(self) -> None:
    """Thread for grabbing the last image in the video stream and putting it in
    the buffer."""

    for frame in self._stream:
      with self._lock:
        self._frame = frame.array
      self._capture.truncate(0)
      if self._stop:
        break

  def close(self) -> None:
    """Joins the thread and closes the stream and the
    :class:`picamera.PiCamera` object."""

    self._stop = True
    self._frame_grabber.join()
    self._capture.close()
    self._cam.close()

  def _get_width(self) -> int:
    return self._cam.resolution[0]

  def _get_height(self) -> int:
    return self._cam.resolution[1]

  def _get_iso(self) -> int:
    return self._cam.iso

  def _get_brightness(self) -> int:
    return self._cam.brightness

  def _get_contrast(self) -> int:
    return self._cam.contrast

  def _get_saturation(self) -> int:
    return self._cam.saturation

  def _get_shutter_speed(self) -> float:
    return self._cam.shutter_speed

  def _get_black_white(self) -> bool:
    return self._cam.color_effects == (128, 128)

  def _get_crop_x_offset(self) -> float:
    return self._cam.zoom[0]

  def _get_crop_y_offset(self) -> float:
    return self._cam.zoom[1]

  def _get_crop_width(self) -> float:
    return self._cam.zoom[2]

  def _get_crop_height(self) -> float:
    return self._cam.zoom[3]

  def _set_width(self, width: float) -> None:
    # The Picamera only accepts width that are multiples of 32
    self._stop_stream()
    self._cam.resolution = (32 * (width // 32), self._get_height())
    self._restart_stream()

  def _set_height(self, height: float) -> None:
    # The Picamera only accepts heights that are multiples of 32
    self._stop_stream()
    self._cam.resolution = (self._get_width(), 32 * (height // 32))
    self._restart_stream()

  def _set_iso(self, iso: float) -> None:
    # The Picamera only accepts a limited range of iso values
    self._cam.iso = min(picamera_iso, key=lambda x: abs(x - iso))

  def _set_brightness(self, brightness: float) -> None:
    self._cam.brightness = brightness

  def _set_contrast(self, contrast: float) -> None:
    self._cam.contrast = contrast

  def _set_saturation(self, saturation: float) -> None:
    self._cam.saturation = saturation

  def _set_shutter_speed(self, shutter_speed: float) -> None:
    self._cam.shutter_speed = shutter_speed

  def _set_black_white(self, boolean: bool) -> None:
    if boolean:
      self._cam.color_effects = (128, 128)
    else:
      self._cam.color_effects = None

  def _set_crop_x_offset(self, x_offset: float) -> None:
    self._cam.zoom = (x_offset, self._get_crop_y_offset(),
                      self._get_crop_width(), self._get_crop_height())

  def _set_crop_y_offset(self, y_offset: float) -> None:
    self._cam.zoom = (self._get_crop_x_offset(), y_offset,
                      self._get_crop_width(), self._get_crop_height())

  def _set_crop_width(self, width: float) -> None:
    self._cam.zoom = (self._get_crop_x_offset(), self._get_crop_y_offset(),
                      width, self._get_crop_height())

  def _set_crop_height(self, height: float) -> None:
    self._cam.zoom = (self._get_crop_x_offset(), self._get_crop_y_offset(),
                      self._get_crop_width(), height)



# coding: utf-8

"""
Very simple test demonstrating the video-extensometry.

To test it, one can draw markers on a paper an move it closer or further to the
camera. This should induce a strain on both `x` and `y` directions.

Required hardware:
  - A camera
"""

import crappy

if __name__ == '__main__':
  graph_extenso = crappy.blocks.Grapher(('t(s)', 'Exx(%)'), ('t(s)', 'Eyy(%)'))

  extenso = crappy.blocks.Video_extenso(camera="Webcam",
                                        raise_on_lost_spot=True,
                                        display_images=True,
                                        white_spots=False)

  crappy.link(extenso, graph_extenso)
  crappy.start()



# coding: utf-8

from time import time
from typing import List, Optional, Tuple

from .block import Block


class PID(Block):
  """A basic implementation of a PID corrector.

  A PID will continuously adjust its output based on the target value and the
  actual measured value, to try to actually reach the target.
  """

  def __init__(self,
               kp: float,
               ki: float = 0,
               kd: float = 0,
               out_max: float = float('inf'),
               out_min: float = -float('inf'),
               target_label: str = 'cmd',
               input_label: str = 'V',
               time_label: str = 't(s)',
               labels: Optional[List[str]] = None,
               reverse: bool = False,
               i_limit: Tuple[Optional[float], Optional[float]] = (None, None),
               send_terms: bool = False,
               freq: float = 500,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      kp: The `P` gain.
      ki: The `I` gain.
      kd: The `D` gain.
      out_max: Ensures the output is always inferior to this value.
      out_min: Ensures the output is always superior to this value.
      target_label: The label carrying the setpoint value.
      input_label: The label carrying the reading of the actual value, to be
        compared with the setpoint.
      time_label: The label carrying the time information in the incoming
        links.
      labels: The two labels that will be sent to downstream blocks. The first
        one is the time label, the second one is the output of the PID. If this
        argument is not given, they default to ``'t(s)'`` and ``'pid'``.
      reverse: If :obj:`True`, reverses the action of the PID.
      i_limit: A :obj:`tuple` containing respectively the lower and upper
        boundaries for the `I` term.
      send_terms: If :obj:`True`, returns the weight of each term in the output
        value. It adds ``'p_term', 'i_term', 'd_term'`` to the output labels.
        This is particularly useful to tweak the gains.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.
    """

    # Attributes of the parent class
    super().__init__()
    self.niceness = -10
    self.freq = freq
    self.verbose = verbose
    self.labels = ['t(s)', 'pid'] if labels is None else labels
    if send_terms:
      self.labels.extend(['p_term', 'i_term', 'd_term'])

    # Setting the gains
    sign = -1 if reverse else 1
    self._kp = sign * kp
    self._ki = sign * kp * ki
    self._kd = sign * kp * kd

    # Setting the limits
    self._out_max = out_max
    self._out_min = out_min
    self._i_min, self._i_max = i_limit

    # Setting the labels
    self._target_label = target_label
    self._input_label = input_label
    self._time_label = time_label

    self._send_terms = send_terms

    # Setting the variables
    self._target = None
    self._last_input = None
    self._last_t = None
    self._i_term = 0

  def begin(self) -> None:
    """Receives the first target and input values and makes sure the given
    labels are correct."""

    # Waiting for the first data to arrive
    data = [link.recv(blocking=True) for link in self.inputs]

    # Getting the first target value
    for dic in data:
      if self._target_label in dic:
        self._target = dic[self._target_label]
        break

    # Getting the first input value
    for dic in data:
      if self._input_label in dic and self._time_label in dic:
        self._last_input = dic[self._input_label]
        self._last_t = dic[self._time_label]
        break

    # Making sure the given labels are correct
    if self._target is None:
      raise IOError(f'No link containing the target label '
                    f'{self._target_label} !')
    if self._last_input is None:
      raise IOError(f'No link containing the input label {self._input_label} '
                    f'and the time label {self._time_label} !')

    # Sending the values to the downstream blocks
    if self._send_terms:
      self.send([self._last_t, 0, 0, 0, 0])
    else:
      self.send([self._last_t, 0])

  def loop(self) -> None:
    """Receives the latest target and input values, calculates the P, I and D
    terms and sends the output to the downstream blocks."""

    # Looping in a non-blocking way
    data = [link.recv_last(blocking=False) for link in self.inputs]

    input_ = None
    t = None

    # Updating the target value
    for dic in data:
      if dic is not None and self._target_label in dic:
        self._target = dic[self._target_label]
        break

    # Getting the latest input value
    for dic in data:
      if dic is not None and self._input_label in dic \
            and self._time_label in dic:
        input_ = dic[self._input_label]
        t = dic[self._time_label]
        break

    # If there's no new input, do nothing
    if input_ is None or t is None:
      return

    delta_t = t - self._last_t
    diff = self._target - input_

    # Calculating the three PID terms
    p_term = self._kp * diff
    self._i_term += self._ki * diff * delta_t
    d_term = - self._kd * (input_ - self._last_input) / delta_t

    self._last_t = t
    self._last_input = input_

    # Clamping the i term if required
    if self._i_min is not None:
      self._i_term = max(self._i_min, self._i_term)
    if self._i_max is not None:
      self._i_term = min(self._i_max, self._i_term)

    # Clamping the output if required
    out = p_term + self._i_term + d_term
    out = min(self._out_max, max(self._out_min, out))

    # Sending the values to the downstream blocks
    if self._send_terms:
      self.send([time() - self.t0, out, p_term, self._i_term, d_term])
    else:
      self.send([time() - self.t0, out])



# coding: utf-8

from time import time
from typing import Optional, List
from dataclasses import dataclass

from .inout import InOut
from .._global import OptionalModule

try:
  from ue9 import UE9
except (ModuleNotFoundError, ImportError):
  UE9 = OptionalModule("LabJackPython")


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Labjack UE9
  channel can have."""

  num: int
  range_num: int = 12
  gain: float = 1
  offset: float = 0
  make_zero: bool = False


class Labjack_ue9(InOut):
  """"""

  def __init__(self,
               channels: List[int],
               gain: Optional[List[float]] = None,
               offset: Optional[List[float]] = None,
               make_zero: Optional[List[bool]] = None,
               resolution: Optional[List[int]] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      channels: A :obj:`list` containing all the channels to read, given as
        :obj:`int`. Only the `AIN` channels can be read by this class, so to
        read the channel `AIN2` the integer `2` should be added to the list.
      gain: A :obj:`list` containing for each channel the gain to apply to the
        measured voltage, as a :obj:`float`. The returned voltage is
        calculated as follows :
        ::

          returned_voltage = gain * measured_voltage + offset

        If not given, no gain is applied to the measured values.
      offset: A :obj:`list` containing for each channel the offset to apply to
        the measured voltage, as a :obj:`float`. The returned voltage is
        calculated as follows :
        ::

          returned_voltage = gain * measured_voltage + offset

        If not given, no offset is applied to the measured values.
      make_zero: A :obj:`list` containing for each channel a :obj:`bool`
        indicating whether the channel should be zeroed or not. If so, data
        will be acquired on this channel before the test starts, and a
        compensation value will be deduced so that the offset of this channel
        is `0`. **It will only take effect if the ``make_zero_delay`` argument
        of the :ref:`IOBlock` controlling the Labjack is set** ! If not given,
        the channels are by default not zeroed.
      resolution: A :obj:`list` containing for each channel the resolution of
        the acquisition as an integer. Refer to Labjack documentation for more
        details. The higher this value the better the resolution, but the lower
        the speed. The possible range is `1` to `12`, and the default is `12`.

    Note:
      All the :obj:`list` given as arguments for the channels should have the
      same length. If that's not the case, all the given lists are treated as
      if they had the same length as the shortest given list.
    """

    super().__init__()

    # Setting the defaults for arguments that are not given
    if resolution is None:
      resolution = [12 for _ in channels]
    if gain is None:
      gain = [1 for _ in channels]
    if offset is None:
      offset = [0 for _ in channels]
    if make_zero is None:
      make_zero = [False for _ in channels]

    self._channels = [_Channel(num=chan, range_num=r_num, gain=g,
                               offset=off, make_zero=make_z)
                      for chan, r_num, g, off, make_z in
                      zip(channels, resolution, gain, offset, make_zero)]

    self._handle = None

  def open(self) -> None:
    """Simply opens the connection to the Labjack."""

    self._handle = UE9()

  def make_zero(self, delay: float) -> None:
    """Overriding of the method of the parent class, because the user can
    choose which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation to
    zero for the channels that shouldn't be zeroed.
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:
        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]

  def get_data(self) -> List[float]:
    """Simply reads sequentially the channels and returns the acquired values,
    corrected by the given gains and offsets."""

    return [time()] + [self._handle.getAIN(chan.num, Resolution=chan.range_num)
                       * chan.gain + chan.offset for chan in self._channels]

  def close(self) -> None:
    """Closes the Labjack if it was already opened."""

    if self._handle is not None:
      self._handle.close()



# coding: utf-8

# Todo:
#  Add downsampling ( self.cam.set_downsampling('XI_DWN_2x2') )
#  Add region of interest selection
#  Add possibility to adjust the timeout

from time import time
from typing import Optional, Tuple
import numpy as np

from .camera import Camera
from .._global import OptionalModule
try:
  from ximea import xiapi
except (ModuleNotFoundError, ImportError):
  xiapi = OptionalModule("ximea", "To use Xiapi cameras, please install the "
                         "official ximea Python module")


class Xiapi(Camera):
  """Camera class for ximeas using official XiAPI."""

  def __init__(self) -> None:
    """"""

    super().__init__()

    self._cam = xiapi.Camera()
    self._img = xiapi.Image()

  def open(self, sn: Optional[str] = None, **kwargs) -> None:
    """Will actually open the camera, args will be set to default unless
    specified otherwise in kwargs.

    If `sn` is given, it will open the camera with the corresponding serial
    number.

    Else, it will open any camera.
    """

    if sn is not None:
      self._cam.open_device_by_SN(sn)
    else:
      self._cam.open_device()

    self.add_scale_setting('width', 1, self._get_w(), self._get_w, self._set_w,
                           self._get_w())
    self.add_scale_setting('height', 1, self._get_h(), self._get_h,
                           self._set_h, self._get_h())
    self.add_scale_setting('xoffset', 0, self._get_w(), self._get_ox,
                           self._set_ox, 0)
    self.add_scale_setting('yoffset', 0, self._get_h(), self._get_oy,
                           self._set_oy, 0)
    self.add_scale_setting('exposure', 28, 500000, self._get_exp,
                           self._set_exp, 10000)
    self.add_scale_setting('gain', 0., 6., self._get_gain, self._set_gain)
    self.add_bool_setting('AEAG', self._get_aeag, self._set_aeag, False)
    self.add_bool_setting('external_trig', self._get_extt,
                          self._set_extt, False)

    self.set_all(**kwargs)
    self._cam.start_acquisition()

  def get_image(self) -> Tuple[float, np.ndarray]:
    """This method get a frame on the selected camera and return a ndarray.

    Returns:
        frame from ximea device (`ndarray height*width`).
    """

    self._cam.get_image(self._img)
    t = time()
    self._frame_nr += 1
    return t, self._img.get_image_data_numpy()

  def close(self) -> None:
    """This method closes properly the camera.

    Returns:
        void return function.
    """

    self._cam.close_device()

  def _get_w(self) -> int:
    return self._cam.get_width()

  def _get_h(self) -> int:
    return self._cam.get_height()

  def _get_ox(self) -> int:
    return self._cam.get_offsetX()

  def _get_oy(self) -> int:
    return self._cam.get_offsetY()

  def _get_gain(self) -> float:
    return self._cam.get_gain()

  def _get_exp(self) -> float:
    return self._cam.get_exposure()

  def _get_aeag(self) -> bool:
    return self._cam.get_param('aeag')

  def _get_extt(self) -> bool:
    r = self._cam.get_trigger_source()
    if r == 'XI_TRG_OFF':
      return False
    else:
      return True

  def _set_w(self, width: int) -> None:
    self._cam.set_width(width)

  def _set_h(self, height: int) -> None:
    self._cam.set_height(height)

  def _set_ox(self, x_offset: int) -> None:
    self._cam.set_offsetX(x_offset)

  def _set_oy(self, y_offset: int) -> None:
    self._cam.set_offsetY(y_offset)

  def _set_gain(self, gain: float) -> None:
    self._cam.set_gain(gain)

  def _set_exp(self, exposure: float) -> None:
    self._cam.set_exposure(exposure)

  def _set_aeag(self, aeag: bool) -> None:
    self._cam.set_param('aeag', int(aeag))

  def _set_extt(self, trig: bool) -> None:
    if trig:
      self._cam.set_gpi_mode('XI_GPI_TRIGGER')
      self._cam.set_trigger_source('XI_TRG_EDGE_RISING')
    else:
      self._cam.set_gpi_mode('XI_GPI_OFF')
      self._cam.set_trigger_source('XI_TRG_OFF')



# coding: utf-8

"""
This program is used to control a biaxial tensile machine.

It demonstrates an equibiaxial cyclic test.
"""

import crappy

path = {'type': 'cyclic', 'value1': 5, 'condition1': 'delay=3',
        'value2': -5, 'condition2': 'delay=3', 'cycles': 0}
path2 = dict(path)
path2['condition1'] = 'delay=5'
path2['condition2'] = 'delay=5'
g1 = crappy.blocks.Generator(path=[path], cmd_label="vx")
g2 = crappy.blocks.Generator(path=[path2], cmd_label="vy")

mot = {'type': 'biaxe',
       'mode': 'speed'
       }

motA = {'port': '/dev/ttyS4',
        'cmd': 'vy',
        }

motB = {'port': '/dev/ttyS5',
        'cmd': 'vx',
        }

motC = {'port': '/dev/ttyS6',
        'cmd': 'vy',
        }

motD = {'port': '/dev/ttyS7',
        'cmd': 'vx',
        }

b = crappy.blocks.Machine([motA, motB, motC, motD], common=mot)
crappy.link(g1, b)
crappy.link(g2, b)
crappy.start()



# encoding: utf-8

"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import find_packages
# To use a consistent encoding
from os import popen, listdir, walk
import sys
from distutils.core import setup, Extension
import platform
with open('crappy/__version__.py') as file:
  for line in file:
    if line.startswith('__version__'):
      __version__ = line.split('=')[1].replace(' ', '').\
        replace('\n', '').replace("'", '')

# Get the long description from the relevant file
with open('docs/source/whatiscrappy.rst', encoding='utf-8') as f:
  long_description = f.read()

v = "%d.%d" % (sys.version_info.major, sys.version_info.minor)
extensions = []
pyFgenModule = clModule = None

if platform.system() == "Linux":
  try:
    # Find the latest runtime version of SiliconSoftware install
    clPath = '/opt/SiliconSoftware/' + \
             sorted(next(walk('/opt/SiliconSoftware/'))[1])[-1] + '/lib64/'
  except StopIteration:
    print("Silicon Software not found, CameraLink will not be supported.")
    # If the software is installed but not found
    # just set clPath manually in this file
    clPath = None
  if clPath:
    clModule = Extension('camera.clModule',
                         sources=['sources/Cl_lib/CameraLink.cpp',
                                  'sources/Cl_lib/pyCameraLink.cpp',
                                  'sources/Cl_lib/clSerial.cpp'],
                         extra_compile_args=["-std=c++11"],
                         extra_link_args=["-l", "python%sm" % v, "-L", clPath,
                                          "-l", "display", "-l", "clsersis",
                                          "-l", "fglib5"],
                         include_dirs=['/usr/local/lib/python%sm/dist-packages'
                                       '/numpy/core/include' % v])
    p = popen("lsmod |grep menable")
    if len(p.read()) != 0:
      print("menable kernel module found, installing CameraLink module.")
      extensions.append(clModule)
    else:
      print("Cannot find menable kernel module, "
            "CameraLink module won't be available.")

if platform.system() == "Windows":
  pyFgenModule = Extension('tool.pyFgenModule',
                           include_dirs=["C:\\python%s\\site-packages\\numpy\\"
                                         "core\\include" % v,
                                         "C:\\Program Files (x86)\\IVI "
                                         "Foundation\\VISA\\WinNT\\include",
                                         "C:\\Program Files\\IVI Foundation\\"
                                         "IVI\\Include"],
                           sources=['sources/niFgen/pyFgen.cpp'],
                           libraries=["niFgen"],
                           library_dirs=["C:\\Program Files\\IVI Foundation\\"
                                         "IVI\\Lib_x64\\msc"],
                           extra_compile_args=["/EHsc", "/WX"])
  if input("would you like to install pyFgen module? ([y]/n)") != "n":
    extensions.append(pyFgenModule)
  clpath = "C:\\Program Files\\SiliconSoftware\\Runtime5.2.1\\"
  clModule = Extension('tool.clModule',
                       include_dirs=[clpath+"include",
                                     "C:\\python{}\\Lib\\site-packages\\numpy"
                                     "\\core\\include".format(
                                       v.replace('.', ''))],
                       sources=['sources/Cl_lib/CameraLink.cpp',
                                'sources/Cl_lib/pyCameraLink.cpp',
                                'sources/Cl_lib/clSerial.cpp'],
                       libraries=["clsersis", "fglib5"],
                       library_dirs=[clpath+"lib\\visualc"],
                       extra_compile_args=["/EHsc", "/WX"])

  p = popen('driverquery /NH |findstr "me4"')
  if len(p.read()) != 0:
    extensions.append(clModule)
  else:
    print("Can't find microEnable4 Device driver, clModule will not be "
          "compiled")

docs_files = [('crappy/' + path, [path + '/' + name for name in paths]) for
              (path, _, paths) in walk('docs/source')]

setup(
  name='crappy',

  version=__version__,

  description='Command and Real-time Acquisition Parallelized in Python',

  long_description=long_description,

  url='https://github.com/LaboratoireMecaniqueLille/crappy',

  project_urls={'Documentation':
                'https://crappy.readthedocs.io/en/latest/index.html'},

  author='LaMcube',
  author_email='antoine.weisrock1@centralelille.fr',

  license='GPL V2',

  zip_safe=False,

  classifiers=['Development Status :: 4 - Beta ',
               'Intended Audience :: Science/Research',
               'Topic :: Software Development :: Build Tools',
               'License :: OSI Approved :: GNU General Public License v2 or '
               'later (GPLv2+)',
               'Programming Language :: Python :: 3.6',
               'Natural Language :: English',
               'Operating System :: OS Independent'],

  keywords='control command acquisition multiprocessing',

  packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

  python_requires=">=3.6",

  ext_package='crappy',
  ext_modules=extensions,

  install_requires=["numpy>=1.21.0"],

  extras_require={'SBC': ['smbus2',
                          'spidev',
                          'Adafruit-Blinka',
                          'Adafruit-ADS1x15',
                          'adafruit-circuitpython-motorkit',
                          'adafruit-circuitpython-mprls',
                          'adafruit-circuitpython-busdevice'],
                  'image': ['opencv-python>=4.0',
                            'Pillow>=8.0.0',
                            'matplotlib>=3.3.0',
                            'SimpleITK>=2.0.0',
                            'scikit-image>=0.18.0'],
                  'hardware': ['pyusb>=1.1.0',
                               'pyserial>=3.4',
                               'pyyaml>=5.3'],
                  'main': ['matplotlib>=3.3.0',
                           'opencv-python>=4.0',
                           'pyserial>=3.4']},

  include_package_data=True,

  data_files=[('crappy/docs', ['docs/Makefile']),

              ('crappy/Examples',
               ['Examples/' + filename for filename in listdir('Examples')
                if filename.endswith('.py')]),

              ('crappy/impact',
               ['impact/' + filename for filename in listdir('impact')
                if filename.endswith('.py')]),

              ('crappy/util',
               ['util/' + filename for filename in listdir('util')
                if not filename.startswith('.')])

              ] + docs_files
)



# coding: utf-8

from time import time
from typing import Union, Dict, Optional

from .path import Path, condition_type


class Ramp(Path):
  """Sends a ramp signal varying linearly over time, until the stop condition
  is met."""

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               condition: Union[str, condition_type],
               speed: float,
               init_value: Optional[float] = None):
    """Sets the args and initializes the parent class.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      condition: The condition for switching to the next path. Refer to
        :ref:`generator path` for more info.
      speed: The slope of the ramp, in `units/s`.
      init_value: If given, overwrites the last value of the signal as the
        starting point for the ramp. In the specific case when this path is the
        first one in the list of dicts, this argument must be given !
    """

    Path.__init__(self, _last_time, _last_cmd)

    if init_value is None and _last_cmd is None:
      raise ValueError('For the first path, an init_value must be given !')

    # Setting the attributes
    self._condition = self.parse_condition(condition)
    self._speed = speed
    self._init_value = init_value if init_value is not None else _last_cmd

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met."""

    # Checking if the stop condition is met
    if self._condition(data):
      raise StopIteration

    # Returning the current value
    return self._init_value + (time() - self.t0) * self._speed



# coding: utf-8

from .block import Block


class Sink(Block):
  """Test block used to get data and do nothing."""

  def __init__(self,
               verbose: bool = False,
               freq: float = 10) -> None:
    """Sets the args and initializes the parent class."""

    super().__init__()
    self.verbose = verbose
    self.freq = freq

  def loop(self) -> None:
    """Simply drops all received data."""

    self.drop()



# coding : utf-8

from time import time, sleep
from re import fullmatch, findall
from typing import Union, List, Optional
from .inout import InOut
from .._global import OptionalModule

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  from spidev import SpiDev
except (ModuleNotFoundError, ImportError):
  SpiDev = OptionalModule("spidev")

# gain
ADS1263_GAIN = {1: 0,
                2: 1,
                4: 2,
                8: 3,
                16: 4,
                32: 5,
                64: 6}

# data rate
ADS1263_DRATE = {2.5: 0x00,
                 5: 0x01,
                 10: 0x02,
                 16.6: 0x03,
                 20: 0x04,
                 50: 0x05,
                 60: 0x06,
                 100: 0x07,
                 400: 0x08,
                 1200: 0x09,
                 2400: 0x0A,
                 4800: 0x0B,
                 7200: 0x0C,
                 14400: 0x0D,
                 19200: 0x0E,
                 38400: 0x0F}

ADS1263_CMD = {'CMD_RESET': 0x06,
               'CMD_START1': 0x08,
               'CMD_STOP1': 0x0A,
               'CMD_START2': 0x0C,
               'CMD_STOP2': 0x0E,
               'CMD_RDATA1': 0x12,
               'CMD_RDATA2': 0x14,
               'CMD_SYOCAL1': 0x16,
               'CMD_SYGCAL1': 0x17,
               'CMD_SFOCAL1': 0x19,
               'CMD_SYOCAL2': 0x1B,
               'CMD_SYGCAL2': 0x1C,
               'CMD_SFOCAL2': 0x1E,
               'CMD_RREG': 0x20,
               'CMD_RREG2': 0x00,
               'CMD_WREG': 0x40,
               'CMD_WREG2': 0x00}

# registration definition
ADS1263_REG = {'REG_ID': 0x00,
               'REG_POWER': 0x01,
               'REG_INTERFACE': 0x02,
               'REG_MODE0': 0x03,
               'REG_MODE1': 0x04,
               'REG_MODE2': 0x05,
               'REG_INPMUX': 0x06,
               'REG_OFCAL0': 0x07,
               'REG_OFCAL1': 0x08,
               'REG_OFCAL2': 0x09,
               'REG_FSCAL0': 0x0A,
               'REG_FSCAL1': 0x0B,
               'REG_FSCAL2': 0x0C,
               'REG_IDACMUX': 0x0D,
               'REG_IDACMAG': 0x0E,
               'REG_REFMUX': 0x0F,
               'REG_TDACP': 0x10,
               'REG_TDACN': 0x11,
               'REG_GPIOCON': 0x12,
               'REG_GPIODIR': 0x13,
               'REG_GPIODAT': 0x14,
               'REG_ADC2CFG': 0x15,
               'REG_ADC2MUX': 0x16,
               'REG_ADC2OFC0': 0x17,
               'REG_ADC2OFC1': 0x18,
               'REG_ADC2FSC0': 0x19,
               'REG_ADC2FSC1': 0x1A}

RST_PIN = 18
CS_PIN = 22
DRDY_PIN = 17

VREF = 5


class Waveshare_high_precision(InOut):
  """This InOut allows acquiring data from Waveshare's High Precision HAT.

  This board features an ADS1263 32-bits ADC, which is what this InOut actually
  drives. The main specificities compared with driving just an ADS1263 are that
  the reference voltage is the board's 5V supply, and that the differential
  acquisition is improved when using pairs of channels (0 & 1, 2 & 3, etc.).

  The sample rate, hardware gain and digital filter can be adjusted. It is also
  possible to acquire data sequentially from several channels.

  The Waveshare HAT is originally meant to be used with a Raspberry Pi, but it
  can be used with any device supporting SPI as long as the wiring is correct
  and the 3.3 and 5V power are supplied.
  """

  def __init__(self,
               spi_port: int = 0,
               gain_hardware: int = 16,
               sample_rate: Union[int, float] = 50,
               channels: Optional[List[str]] = None,
               digital_filter: int = 4,
               gain: float = 1,
               offset: float = 0) -> None:
    """Sets the args and initializes the parent class.

    Args:
      spi_port: The SPI port for communicating with the Waveshare HAT.
      gain_hardware: A programmable gain for the signal. Should be one of :
        ::

          1, 2, 4, 8, 16, 32

        Allows increasing the resolution of the ADC, but also divides the full
        scale range and greatly reduces the noise performance when set greater
        than 8. This value applies to all the channels.
      sample_rate: The number of samples per second to acquire. Should be one
        of:
        ::

          2.5, 5, 10, 16?6, 20, 50, 60, 100, 400, 1200, 2400, 4800, 7200,
          14400, 19200, 38400

        The actual achieved sample rate might be lower depending on the
        capability of the PC and the load on the processor. Note that the
        greater the sample rate, the greater the noise. For multiple channels,
        the achieved sample rate is roughly the target sample rate divided by
        the number of channels.
      channels: A :obj:`list` containing strings representing the channels to
        acquire. Each channel must follow on of the two syntax :
        ::

          'INi', i in range(10)

        or else
        ::

          'INi-INj', i, j in range(10)

        With the first syntax, the output is the voltage difference between the
        acquired channel and the ground of the power supply. With the second
        syntax, the output value is simply the difference between the voltages
        of the channels `i` and `j`. It is preferable to use the channels in
        pair (0 & 1, 2 & 3, etc.) for differential acquisition. The data from
        the different channels is acquired sequentially, not all at once.
      digital_filter: The Waveshare Hat features a digital filter that can
        accept different settings. Refer to the documentation of the ADS1263
        for more detail.
      gain: Allows to tune the output values of the ADC according to the
        formula:
        ::

          output = gain * tension + offset.

        The same gain applies to all the channels.
      offset: Allows to tune the output values of the ADC according to the
        formula:
        ::

          output = gain * tension + offset.

        The same offset applies to all the channels.

    Important:
      When the ``gain_hardware`` is greater than 1, the PGA cannot amplify
      above 4.7V or under 0.3V. For example a 2.8V signal read with a gain of 2
      would be read as 4.7V after the PGA, not 4.8V ! Beware !
    """

    super().__init__()
    self._gain = gain
    self._offset = offset
    self._bus = SpiDev(spi_port, 0)

    # Checking the validity of the arguments
    if gain_hardware not in ADS1263_GAIN:
      raise ValueError(f"gain should be in {list(ADS1263_GAIN)}")
    self._gain_hardware = gain_hardware

    if sample_rate not in ADS1263_DRATE:
      raise ValueError(f'sample rate should be in {list(ADS1263_DRATE)}')
    self._sample_rate = sample_rate

    if digital_filter not in range(5):
      raise ValueError(f'digital_filter should be in {list(range(5))}')
    self._filter = digital_filter

    # Parsing the channels to check the right syntax was given
    if channels is None:
      channels = ['IN0']

    for channel in channels:
      if fullmatch(r'IN\d', channel) is None and \
            fullmatch(r'IN\d\s?-\s?IN\d', channel) is None:
        raise ValueError("Valid formats for adc_channels values are "
                         "either 'INi' (i in range(10)) or 'INi - INj'")
    self._chan = channels

  def open(self) -> None:
    """Sets up the GPIO and the different parameters on the ADS1263."""

    # Setting up the GPIO
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(RST_PIN, GPIO.OUT)
    GPIO.setup(CS_PIN, GPIO.OUT)
    GPIO.setup(DRDY_PIN, GPIO.IN, pull_up_down=GPIO.PUD_UP)

    # Setting up the SPI bus
    self._bus.max_speed_hz = 2000000
    self._bus.mode = 0b01

    # Resetting the ADS1263
    GPIO.output(RST_PIN, GPIO.HIGH)
    sleep(0.2)
    GPIO.output(RST_PIN, GPIO.LOW)
    sleep(0.2)
    GPIO.output(RST_PIN, GPIO.HIGH)
    sleep(0.2)

    # Un-raising the reset flag
    self._write_reg(ADS1263_REG['REG_POWER'], 0x01)

    # Applying the sample rate and gain settings
    self._write_cmd(ADS1263_CMD['CMD_STOP1'])
    self._write_reg(ADS1263_REG['REG_MODE2'],
                    ADS1263_DRATE[self._sample_rate] |
                    ADS1263_GAIN[self._gain_hardware] << 4)

    # Setting the 5V supply as voltage reference
    self._write_reg(ADS1263_REG['REG_REFMUX'], 0x24)
    # Setting the first channel
    self._set_channel(self._chan[0])
    # Setting the filter
    self._write_reg(ADS1263_REG['REG_MODE1'], self._filter << 5)
    # Disabling the status and checksum bits
    self._write_reg(ADS1263_REG['REG_INTERFACE'], 0x00)

    # Starting the acquisition
    self._write_cmd(ADS1263_CMD['CMD_START1'])

  def get_data(self) -> List[float]:
    """Reads the channels sequentially, and returns all the values along with a
    timestamp.

    Time is returned first, and the values are then returned in the same order
    as the channels were given.
    """

    out = [time()]

    # Reading the channels sequentially
    for channel in self._chan:

      # Don't bother setting the channel if it never changes !
      if len(self._chan) > 1:
        self._set_channel(channel)

      # Waiting for a conversion to complete
      self._wait_drdy()

      # Reading the data from the buffer
      GPIO.output(CS_PIN, GPIO.LOW)
      self._bus.writebytes([ADS1263_CMD['CMD_RDATA1']])
      buf = self._bus.readbytes(4)
      GPIO.output(CS_PIN, GPIO.HIGH)

      # Assembling the data from the four registers
      read = (buf[0] << 24) & 0xFF000000
      read |= (buf[1] << 16) & 0xFF0000
      read |= (buf[2] << 8) & 0xFF00
      read |= (buf[3]) & 0xFF

      # Converting to Volts
      if read >> 31 & 0b1:
        volt = (read / 0x80000000 - 2) * VREF / self._gain_hardware
      else:
        volt = read / 0x7FFFFFFF * VREF / self._gain_hardware

      out.append(self._gain * volt + self._offset)

    return out

  def close(self) -> None:
    """Closes the SPI bus and resets the GPIOs."""

    self._bus.close()
    GPIO.cleanup()

  def _write_cmd(self, cmd: int) -> None:
    """Writes a command to the ADS1263 and manages the CS pin."""

    GPIO.output(CS_PIN, GPIO.LOW)
    self._bus.writebytes([cmd])
    GPIO.output(CS_PIN, GPIO.HIGH)

  @staticmethod
  def _wait_drdy() -> None:
    """Waits for the DRDY_PIN to be low and returns, or raises a
    :exc:`TimeoutError` if data wasn't available within 0.5s."""

    t0 = time()
    while time() - t0 < 0.5:
      if not GPIO.input(DRDY_PIN):
        return
    raise TimeoutError

  def _write_reg(self, reg: int, data: int) -> None:
    """Writes data to a register of the ADS1263 and manages the CS pin."""

    GPIO.output(CS_PIN, GPIO.LOW)
    self._bus.writebytes([ADS1263_CMD['CMD_WREG'] | reg, 0x00, data])
    GPIO.output(CS_PIN, GPIO.HIGH)

  def _set_channel(self, channel: str) -> None:
    """Parses the channel string to set, and writes the corresponding values to
    the ADS1263."""

    # Parsing the channel
    try:
      pos_chan, neg_chan = findall(r'\d', channel)
    except ValueError:
      pos_chan, *_ = findall(r'\d', channel)
      neg_chan = 10

    # Setting it
    self._write_reg(ADS1263_REG['REG_INPMUX'],
                    int(pos_chan) << 4 | int(neg_chan))



# coding: utf-8

"""
Basic example demonstrating the use of the DISCorrel block.

Required hardware:
  - Any camera
"""

import crappy

if __name__ == "__main__":
  dis = crappy.blocks.DISCorrel('Webcam',
                                fields=['x', 'y'],
                                labels=['t(s)', 'x(pix)', 'y(pix)'],
                                display_images=True)

  graph = crappy.blocks.Grapher(('x(pix)', 'y(pix)'))
  crappy.link(dis, graph)
  crappy.start()



# coding: utf-8

from multiprocessing import Pipe
from time import time
from threading import Thread
from copy import copy
from typing import Callable, Union, Any, Dict, Optional, List

from .._global import CrappyStop
from ..modifier import Modifier


class Link:
  """This class is used for transferring information between the blocks.

  The created link is unidirectional, from the input block to the output block.
  Under the hood, a link is basically a :class:`multiprocessing.Pipe` with
  extra features.

  Note:
    You can add one or multiple :ref:`Modifiers` to modify the transferred
    value. The modifiers should either be children of :ref:`Modifier` or
    callables taking a :obj:`dict` as argument and returning a :obj:`dict`.
  """

  count = 0

  def __init__(self,
               input_block=None,
               output_block=None,
               conditions: List[Union[Callable, Modifier]] = None,
               modifiers: List[Union[Callable, Modifier]] = None,
               timeout: float = 1,
               action: str = "warn",
               name: Optional[str] = None) -> None:
    """Sets the instance attributes.

    Args:
      input_block: The Block sending data through the link.
      output_block: The Block receiving data through the link.
      conditions: Deprecated, kept only for backward-compatibility.
      modifiers: A :obj:`list` containing children of :ref:`Modifier` and/or
        callables. If several objects given ,they will be called in the given
        order. See :ref:`Modifiers` for more information.
      timeout: Sets a timeout for sending the data in the link.
      action: Action to perform in case of a :exc:`TimeoutError` during the
        :meth:`send` method. Should be in:
        ::

          'warn', 'kill', 'NoWarn',

        any other value would be for debugging.
      name: Name of the link, to differentiate it from the others when
        debugging. If no specific name is given, the links are anyway numbered
        in the order in which they are instantiated in the code.
    """

    # For compatibility (condition is deprecated, use modifier)
    if conditions is not None:
      if modifiers is not None:
        modifiers += conditions
      else:
        modifiers = conditions

    # Setting the attributes
    count = self._count_links()
    self.name = name if name is not None else f'link{count}'
    self._in, self._out = Pipe()
    self._modifiers = modifiers
    self._action = action
    self._timeout = timeout

    # Associating the link with the input and output blocks if they are given
    if input_block is not None and output_block is not None:
      input_block.add_output(self)
      output_block.add_input(self)

  @classmethod
  def _count_links(cls) -> int:
    """Simply increments the link count and returns it."""

    cls.count += 1
    return cls.count

  def send(self, value: Union[Dict[str, Any], str]) -> None:
    """Sends a value through the link.

    In case of a timeout exception, executes the user-defined action. Raises
    any other exception caught.
    """

    # Trying to send a value through a link
    try:
      send_job = Thread(target=self._send_timeout, args=(value,), daemon=True)
      send_job.start()

      # Waits for the thread to return
      send_job.join(self._timeout)

      # If it's taking too long, raising an exception
      if send_job.is_alive():
        raise TimeoutError

    # If a timeout exception is raised, handling it according to the action
    except TimeoutError as exc:
      # Warning the user
      if self._action == "warn":
        print(f"WARNING : Timeout error in pipe send! Link name: {self.name}")
      # Stopping the program
      elif self._action == "kill":
        raise TimeoutError("Killing Link : " + str(exc))
      # Simply ignoring
      elif self._action == "NoWarn":
        pass
      # Printing the user-defined message
      else:
        print(self._action)

    # Raising any other caught exception
    except Exception as exc:
      print(f"Exception in link send {self.name} : {str(exc)}")
      raise

  def _send_timeout(self, value: Union[Dict[str, Any], str]) -> None:
    """Method for sending data with a given timeout on the link."""

    try:
      # Sending if the value is None or a string
      if self._modifiers is None or isinstance(value, str):
        self._out.send(value)

      # Else, first applying the modifiers
      else:
        for mod in self._modifiers:
          # Case when the modifier is a class
          if hasattr(mod, 'evaluate'):
            value = mod.evaluate(copy(value))
          # Case when the modifier is a method
          else:
            value = mod(copy(value))
          # Exiting the for loop if nothing left in the dict to send
          if value is None:
            break

        # Finally, sending the dict to the link
        if value is not None:
          self._out.send(value)

    # Raising any exception caught, but first sending a stop message downstream
    except Exception as exc:
      print(f"Exception in link {self.name} : {str(exc)}")
      if not self._out.closed:
        self._out.send('close')
        self._out.close()
      raise

  def recv(self, blocking: bool = True) -> Optional[Dict[str, Any]]:
    """Receives data from a link and returns it as a dict.

    Note:
      If ``blocking`` is :obj:`False`, returns :obj:`None` if there is no
      data.

    Args:
      blocking (:obj:`bool`, optional): Enables (:obj:`True`) or disables
        (:obj:`False`) blocking mode.

    Returns:
      If ``blocking`` is :obj:`True`, :meth:`recv` method will wait until data
      is available in the pipe and return the received data. Otherwise, it will
      check if there is data available to return, or return :obj:`None` if the
      pipe is empty.
    """

    try:
      if blocking or self.poll():
        # Simply collecting the data to receive
        ret = self._in.recv()

        # Raising a CrappyStop in case a string is received
        if isinstance(ret, str):
          raise CrappyStop

        return ret

    # If a timeout exception is raised, handling it according to the action
    except TimeoutError as exc:
      # Warning the user
      if self._action == "warn":
        print(f"WARNING : Timeout error in pipe recv! Link name: {self.name}")
      # Stopping the program
      elif self._action == "kill":
        raise TimeoutError(f"Killing Link : {exc}")
      # Simply ignoring
      elif self._action == "NoWarn":
        pass
      # Printing the user-defined message
      else:
        print(self._action)

    # Raising any CrappyStop
    except CrappyStop:
      raise

    # Raising any other caught exception and displaying message
    except Exception as exc:
      print(f"Exception in link recv {self.name} : {str(exc)}")
      raise

  def poll(self) -> bool:
    """Simple wrapper telling whether there's data in the link or not."""

    return self._in.poll()

  def clear(self) -> None:
    """Flushes the link."""

    while self.poll():
      self._in.recv_bytes()

  def recv_last(self, blocking: bool = False) -> Optional[Dict[str, Any]]:
    """Returns only the last value in the pipe, dropping all the others.

    Note:
      If ``blocking`` is :obj:`False`, will return :obj:`None` if there is no
      data waiting.

      If ``blocking`` is :obj:`True`, will wait for at least one data.

    Warning:
      Unlike :meth:`recv`, default is non blocking.
    """

    # First, block if necessary
    data = self.recv(blocking)
    if data is None:
      return

    # Then, flush the pipe and keep only the last value
    while True:
      new = self.recv(blocking=False)
      if new is None:
        return data
      data = new

  def recv_chunk(self,
                 blocking: bool = True) -> Optional[Dict[str, List[Any]]]:
    """Returns all the data waiting in a link.

    Note:
      If ``blocking`` is :obj:`False`, will return :obj:`None` if there is no
      data waiting.

      If ``blocking`` is :obj:`True`, will wait for at least one data.
    """

    # First, block if necessary and return if the link is empty
    ret = self.recv(blocking)
    if ret is None:
      return

    # Putting the received values in lists
    for label, value in ret.items():
      ret[label] = [value]

    while True:
      try:
        data = self.recv(blocking=False)

      # Sending a stop message if a CrappyStop is raised
      except CrappyStop:
        self._out.send("stop")
        return ret

      # Return when the link is empty
      if data is None:
        return ret

      # Adding the received data to the created lists
      for label in ret:
        try:
          ret[label].append(data[label])
        # Raising an exception in case a label is missing
        except KeyError:
          raise IOError(f"{str(self)} Got data without label {label}")

  def recv_delay(self, delay: float) -> Dict[str, List[Any]]:
    """Same as :meth:`recv_chunk` except it runs for a given delay no matter
    if the link is empty or not.

    Useful for blocks with a low looping frequency that don't need data so
    frequently.

    Note:
      All the :meth:`recv` calls are blocking so this method will take at least
      ``delay`` seconds to return, but it could be more since it may wait for
      data.

      Also, it will return at least one reading.
    """

    t_init = time()
    # This first call to recv is blocking
    ret = self.recv(blocking=True)

    # Putting the received values in lists
    for label, value in ret.items():
      ret[label] = [value]

    while time() - t_init < delay:
      try:
        data = self.recv(blocking=True)

      # Sending a stop message if a CrappyStop is raised
      except CrappyStop:
        self._out.send("stop")
        break

      # Adding the received data to the created lists
      for label in ret:
        try:
          ret[label].append(data[label])
        # Raising an exception in case a label is missing
        except KeyError:
          raise IOError(f"{str(self)} Got data without label {label}")

    return ret

  def recv_chunk_no_stop(self) -> Optional[Dict[str, List[Any]]]:
    """Experimental feature, to be used in :meth:`finish` methods to recover
    the final remaining data (possibly after a stop signal)."""

    # First, collecting all the remaining data
    recv = []
    while self.poll():
      data = self._in.recv()
      if isinstance(data, dict):
        recv.append(data)

    # Then, organizing it into a nice dict to return
    ret = {}
    for data in recv:
      for label, value in data.items():
        if label in ret:
          ret[label].append(value)
        else:
          ret[label] = [value]

    return ret if ret else None


def link(in_block,
         out_block,
         condition: Optional[Union[List[Union[Modifier, Callable]],
                                   Union[Modifier, Callable]]] = None,
         modifier: Optional[Union[List[Union[Modifier, Callable]],
                                  Union[Modifier, Callable]]] = None,
         timeout: float = 1,
         action: str = "warn",
         name: Optional[str] = None) -> None:
  """Function linking two blocks, allowing to send data from one to the other.

  The created link is unidirectional, from the input block to the output block.
  Under the hood, a link is basically a :class:`multiprocessing.Pipe` with
  extra features.

  Args:
    in_block: The Block sending data through the link.
    out_block: The Block receiving data through the link.
    condition: Deprecated, kept only for backward-compatibility.
    modifier: Either a child class of :ref:`Modifier`, or a callable, or a
      :obj:`list` containing such objects. If several given (in a list), calls
      them in the  given order. See :ref:`Modifiers` for more information.
    timeout: Sets a timeout for sending the data in the link.
    action: Action to perform in case of a :exc:`TimeoutError` during the
      :meth:`send` method. Should be in:
      ::

        'warn', 'kill', 'NoWarn',

      any other value would be for debugging.
    name: Name of the link, to differentiate it from the others when debugging.
      If no specific name is given, the links are anyway numbered in the order
      in which they are instantiated in the code.
  """

  # Forcing the conditions and modifiers into lists
  if condition is not None and not isinstance(condition, list):
    condition = [condition]
  if modifier is not None and not isinstance(modifier, list):
    modifier = [modifier]

  # Actually creating the Link object
  Link(input_block=in_block,
       output_block=out_block,
       conditions=condition,
       modifiers=modifier,
       timeout=timeout,
       action=action,
       name=name)



# coding: utf-8

"""
Program used to control the solidification furnace.

It uses a Labjack T7 to send PWN signals to the transistors controlling the
heating element of each section of the furnace. The temperature of each section
is measured using a thermocouple.
"""

import crappy

SHOW_PID = 3

MED = 20
MEAN = 50

FREQ = 100
P = .25
I = .03
D = 20

pins = [0, 2, 3, 4, 5]

T = 200

v = dict()
v[0] = T
v[2] = T + 10
v[3] = T
v[4] = T
v[5] = T


class dc_to_clk(crappy.modifier.Modifier):
  def __init__(self, lbl):
    super().__init__()
    self.lbl = lbl

  def evaluate(self, data):
    # data[self.l] = int(data[self.l] * 80000000 / FREQ)
    if data[self.lbl] < 0.01:
      cmd = 0
    elif data[self.lbl] > .99:
      cmd = 1
    else:
      cmd = data[self.lbl]
    data[self.lbl] = int((1 - cmd) * 80000000 / FREQ)
    return data


clock_config = [
                ('DIO_EF_CLOCK0_ENABLE', 0),
                ('DIO_EF_CLOCK0_DIVISOR', 1),
                ('DIO_EF_CLOCK0_ROLL_VALUE', int(80000000/FREQ)),
                ('DIO_EF_CLOCK0_ENABLE', 1),
                ]


def pwm_config(i):
  return [
      ('DIO%d_EF_ENABLE' % i, 0),
      ('DIO%d_EF_INDEX' % i, 0),
      ('DIO%d_EF_OPTIONS' % i, 0),
      ('DIO%d_EF_CONFIG_A' % i, int(80000000 / FREQ * .5)),
      ('DIO%d_EF_ENABLE' % i, 1),
    ]


# g = crappy.blocks.Generator([dict(type='constant', condition=None,
# value=200)])
pwm_chan = [dict(name="DIO%d_EF_CONFIG_A" % i,
                 direction=1, write_at_open=pwm_config(i)) for i in pins]
# Adding the clock config to the first chan
pwm_chan[0]['write_at_open'][0:0] = clock_config

th_chan = [dict(name='AIN%d' % i, thermocouple='K') for i in pins]


lj = crappy.blocks.IOBlock("Labjack_t7", channels=pwm_chan+th_chan,
                           labels=['t(s)']+['T%d' % i for i in pins],
                           cmd_labels=['pwm%d' % i for i in pins],
                           verbose=True)

pid_list = []
gen_list = []
graph_cmd = crappy.blocks.Grapher(*[('t(s)', 'pwm%d' % i) for i in pins])
for i in pins:
  pid_list.append(crappy.blocks.PID(P, I if i != 5 else 0, D,
                                    input_label='T%d' % i,
                                    out_max=1, out_min=0,
                                    i_limit=(0.5, -0.5),
                                    send_terms=(SHOW_PID is not None
                                                and i == SHOW_PID),
                                    labels=['t(s)', 'pwm%d' % i]))

  gen_list.append(crappy.blocks.Generator(
    [dict(type='constant', condition=None, value=v[i])]))

  crappy.link(gen_list[-1], pid_list[-1])
  crappy.link(pid_list[-1], lj, modifier=[dc_to_clk('pwm%d' % i)])
  crappy.link(lj, pid_list[-1],
              modifier=[crappy.modifier.Median(MED),
                        crappy.modifier.Moving_avg(MEAN)])
  crappy.link(pid_list[-1], graph_cmd)

graph = crappy.blocks.Grapher(*[('t(s)', 'T%d' % i) for i in pins])
crappy.link(lj, graph,
            modifier=[crappy.modifier.Median(MED),
                      crappy.modifier.Moving_avg(MEAN)])


if SHOW_PID:
  graph_pid = crappy.blocks.Grapher(
      ('t(s)', 'p_term'),
      ('t(s)', 'i_term'),
      ('t(s)', 'd_term'),
      ('t(s)', 'pwm%d' % SHOW_PID))
  crappy.link(pid_list[pins.index(SHOW_PID)], graph_pid)

crappy.start()



# coding: utf-8

import crappy

gx = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F3>5',
                               'value1': 10, 'condition1': 'F3<-5',
                               'value2': -10}], cmd_label='vx')

gy = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F1>5',
                               'value1': 10, 'condition1': 'F1<-5',
                               'value2': -10}], cmd_label='vy')

m1 = {'port': '/dev/ttyS4', 'cmd': 'vy'}
m2 = {'port': '/dev/ttyS5', 'cmd': 'vy'}
m3 = {'port': '/dev/ttyS6', 'cmd': 'vx'}
m4 = {'port': '/dev/ttyS7', 'cmd': 'vx'}
common = {'type': 'biaxe'}

m = crappy.blocks.Machine([m1, m2, m3, m4], common)
s = crappy.blocks.IOBlock('Comedi', channels=[1, 3], gain=[3749, 3749],
                          labels=['t(s)', 'F1', 'F3'])

g = crappy.blocks.Grapher(('t(s)', 'F1'), ('t(s)', 'F3'))

crappy.link(s, g)
crappy.link(s, gx)
crappy.link(s, gy)
crappy.link(gx, m)
crappy.link(gy, m)

crappy.start()



# coding:utf-8

from .constant import Constant
from .cyclic import Cyclic
from .ramp import Ramp
from .cyclic_ramp import Cyclic_ramp
from .inertia import Inertia
from .sine import Sine
from .protection import Protection
from .custom import Custom



# coding: utf-8

from typing import Union, List, Optional
from .block import Block
from ..inout import inout_dict


class IOBlock(Block):
  """This block is meant to drive :ref:`In / Out` objects. It can acquire data,
  and/or set commands. One IOBlock can only drive a single InOut.

  If it has incoming links, it will set the commands received over the labels
  given in ``cmd_labels``. Additional commands to set at the very beginning or
  the very end of the test can also be specified.

  If it has outgoing links, it will acquire data and send it downstream over
  the labels given in ``labels``. It is possible to trigger the acquisition
  using a predefined label.
  """

  def __init__(self,
               name: str,
               labels: Optional[List[str]] = None,
               cmd_labels: Optional[List[str]] = None,
               trigger_label: Optional[str] = None,
               streamer: bool = False,
               initial_cmd: Optional[Union[list]] = None,
               exit_cmd: Optional[list] = None,
               make_zero_delay: Optional[float] = None,
               spam: bool = False,
               freq: Optional[float] = None,
               verbose: bool = False,
               **kwargs) -> None:
    """Sets the args and initializes the parent class.

    Args:
      name: The name of the :ref:`In / Out` class to instantiate.
      labels: A :obj:`list` containing the output labels for InOuts that
        acquire data. They correspond to the values returned by the InOut's
        :meth:`get_data` method, so there should be as many labels as values
        returned, and given in the appropriate order. The first label must
        always be the timestamp, preferably called ``'t(s)'``. This argument
        can be omitted if :meth:`get_data` returns a :obj:`dict` instead of a
        :obj:`list`. Ignored if the block has no output link.
      cmd_labels: A :obj:`list` of the labels considered as inputs for this
        block, for InOuts that set commands. The values received from these
        labels will be passed to the InOut's :meth:`set_cmd` method, in the
        same order as the labels are given. Usually, time is not part of the
        cmd_labels. Ignored if the block has no input link.
      trigger_label: If given, the block will only read data whenever a value
        (can be any value) is received on this label. Ignored if the block has
        no output link. A trigger label can also be a cmd label.
      streamer: If :obj:`False`, the :meth:`get_data` method of the InOut
        object is called for acquiring data, else it's the :meth:`get_stream`
        method.
      initial_cmd: An initial command for the InOut, set during
        :meth:`prepare`. If given, there must be as many values as in
        cmd_labels.
      exit_cmd: A final command for the InOut, set during :obj:`finish`. If
        given, there must be as many values as in cmd_labels.
      make_zero_delay: If set, will acquire data before the beginning of the
        test and use it to offset all the labels to zero. The data will be
        acquired during the given number of seconds. Ignored if the block has
        no output links.
      spam: If :obj:`False`, the block will call :meth:`set_cmd` on the
        InOut object only if the current command is different from the
        previous.
      freq: The block will try to loop as this frequency, or as fast as
        possible if no value is given.
      verbose: If :obj:`True`, prints the looping frequency of the block.
      **kwargs: The arguments to be passed to the :ref:`In / Out` class.
    """

    super().__init__()
    self.niceness = -10
    if freq is not None:
      self.freq = freq
    self.verbose = verbose

    # The label argument can be omitted for streaming
    if labels is None and streamer:
      self.labels = ['t(s)', 'stream']
    else:
      self.labels = labels

    # The labels to get
    self._cmd_labels = cmd_labels
    self._trig_label = trigger_label

    # Checking that the initial_cmd and exit_cmd length are consistent
    if cmd_labels is not None:
      if initial_cmd is not None and len(initial_cmd) != len(cmd_labels):
        raise ValueError("There should be as many values in initial_cmd as "
                         "there are in cmd_labels !")
      if exit_cmd is not None and len(exit_cmd) != len(cmd_labels):
        raise ValueError("There should be as many values in exit_cmd as "
                         "there are in cmd_labels !")

    self._initial_cmd = initial_cmd
    self._exit_cmd = exit_cmd

    self._streamer = streamer
    self._spam = spam
    self._make_zero_delay = make_zero_delay

    self._stream_started = False
    self._last_cmd = None
    self._prev_value = dict()

    # Instantiating the device
    self._device = inout_dict[name.capitalize()](**kwargs)

  def prepare(self) -> None:
    """Checks the consistency of the link layout, opens the device and sets the
    initial command if required."""

    # Checking that the block has inputs or outputs
    if not self.inputs and not self.outputs:
      raise IOError('Error ! The IOBlock is neither an input nor an output !')

    # cmd_labels must be defined when the block has inputs
    if self.inputs and self._cmd_labels is None:
      raise ValueError('Error ! The IOBlock has incoming links but no '
                       'cmd_labels have been given !')

    self._read = bool(self.outputs)
    self._write = bool(self.inputs)

    # Now opening the device
    self._device.open()

    # Acquiring data for offsetting the output
    if self._read and self._make_zero_delay is not None:
      self._device.make_zero(self._make_zero_delay)

    # Writing the first command before the beginning of the test if required
    if self._write and self._initial_cmd is not None:
      self._device.set_cmd(*self._initial_cmd)
      self._last_cmd = self._initial_cmd

  def loop(self) -> None:
    """Gets the latest command, reads data from the device and sets the
    command.

    Also handles the trig label if one was given, and manages the buffer for
    the previously received commands.
    """

    # Receiving all the latest data waiting in the links
    # Cannot use self.get_last because trig_label needs a special handling
    recv_data = dict()
    for link in self.inputs:
      latest = link.recv_last(blocking=False)
      if latest is not None:
        recv_data.update(latest)

    # Reading data from the device if there's no trig_label or if data has been
    # received on this trig_label
    if self._read:
      if self._trig_label is None or self._trig_label in recv_data:
        self._read_data()

    # Storing the latest received values in the buffer
    self._prev_value.update(recv_data)
    # Completing the values that may be missing in recv_data
    recv_data.update(self._prev_value)

    if self._write:
      # At the very beginning of the test, there may not be any received value
      if not recv_data:
        return

      # Keeping only the labels in cmd_labels
      cmd = [val for label, val in recv_data.items()
             if label in self._cmd_labels]

      # If not all cmd_labels have a value, returning without calling set_cmd
      if len(cmd) != len(self._cmd_labels):
        print(f"WARNING ! Not enough values received in the "
              f"{type(self._device).__name__} InOut to set the cmd, cmd "
              f"not set !")
        return

      # Setting the command if it's different from the previous or spam is True
      if cmd != self._last_cmd or self._spam:
        self._device.set_cmd(*cmd)
        self._last_cmd = cmd

  def finish(self) -> None:
    """Stops the stream, sets the exit command if necessary, and closes the
    device."""

    # Stopping the stream
    if self._streamer:
      self._device.stop_stream()

    # Setting the exit command
    if self._write and self._exit_cmd is not None:
      self._device.set_cmd(*self._exit_cmd)

    # Closing the device
    self._device.close()

  def _read_data(self) -> None:
    """Reads the data or the stream, offsets the timestamp and sends the data
    to downstream blocks."""

    if self._streamer:
      # Starting the stream if needed
      if not self._stream_started:
        self._device.start_stream()
        self._stream_started = True
      # Actually getting the stream
      data = self._device.return_stream()
    else:
      # Regular reading of data
      data = self._device.return_data()

    if data is None:
      return

    # Making time relative to the beginning of the test
    if isinstance(data, dict) and 't(s)' in data:
      data['t(s)'] -= self.t0
    else:
      data[0] -= self.t0

    self.send(data)



# coding: utf-8

"""Example showing the use of the UController block.

This example is meant to be used in combination with the microcontroller.py or
microcontroller.ino templates in the tools of Crappy.

Required hardware:
  - A microcontroller

"""

import crappy

if __name__ == "__main__":

  gen = crappy.blocks.Generator([{'type': 'constant',
                                  'value': i,
                                  'condition': 'delay=10'}
                                 for i in range(1, 6)],
                                cmd_label='freq',
                                freq=50)

  uc = crappy.blocks.UController(cmd_labels=['freq'], labels=['nr'],
                                 init_output={'nr': 0}, freq=50)

  graph = crappy.blocks.Dashboard(labels=['t(s)', 'nr'])

  crappy.link(gen, uc)
  crappy.link(uc, graph)

  crappy.start()



# coding: utf-8

from time import time

from .actuator import Actuator


class Fake_motor(Actuator):
  """To run test programs without a physical actuator.

  Note:
    A virtual motor driven by a voltage, you can set its properties with the
    args. It has the same methods as a real motor: :meth:`open`,
    :meth:`set_speed`, :meth:`get_speed`, :meth:`get_position`.
  """

  def __init__(self,
               inertia: float = 0.5,
               torque: float = 0.,
               kv: float = 1000,
               rv: float = 0.4,
               fv: float = 2e-5,
               sim_speed: float = 1,
               initial_speed: float = 0,
               initial_pos: float = 0,
               **kwargs) -> None:
    """Sets the instance attributes.

    Args:
      inertia (:obj:`float`, optional): Inertia of the motor (`kg.m²`).
      torque(:obj:`float`, optional): A torque applied on the axis (`N.m`).
      kv(:obj:`float`, optional): The electrical constant of the motor
        (`t/min/V`).
      rv(:obj:`float`, optional): The solid friction.
      fv(:obj:`float`, optional): The fluid friction.
      sim_speed(:obj:`float`, optional): Speed factor of the simulation.
      initial_speed(:obj:`float`, optional): (`rpm`)
      initial_pos(:obj:`float`, optional): (turns)
    """

    super().__init__()
    self.inertia = inertia
    self.torque = torque
    self.kv = kv
    self.rv = rv
    self.fv = fv
    self.sim_speed = sim_speed
    self.initial_speed = initial_speed
    self.initial_pos = initial_pos
    assert not kwargs, "Fake_motor got invalid kwarg(s): " + str(kwargs)

  def open(self) -> None:
    self.rpm = self.initial_speed
    self.pos = self.initial_pos
    self.u = 0  # V
    self.t = time() * self.sim_speed

  def stop(self) -> None:
    self.set_speed(0)

  def close(self) -> None:
    pass

  def update(self) -> None:
    """Updates the motor rpm.

    Note:
      Supposes `u` is constant for the interval `dt`.
    """

    t1 = time() * self.sim_speed
    dt = (t1 - self.t)
    self.t = t1
    f = self.u * self.kv - self.torque - self.rpm * \
        (1 + self.rv + self.rpm * self.fv)
    drpm = f / self.inertia * dt
    self.pos += dt * (self.rpm + drpm / 2)
    self.rpm += drpm

  def get_speed(self) -> float:
    """Return the motor speed (rpm)."""

    self.update()
    return self.rpm

  def get_position(self) -> float:
    """Returns the motor position."""

    self.update()
    return self.pos

  def set_speed(self, u: float) -> None:
    """Sets the motor `cmd` in volts."""

    self.update()
    self.u = u



# coding: utf-8

from .actuator import MetaActuator, Actuator
from .biaxe import Biaxe
from .biotens import Biotens
from .cmDrive import CM_drive
from .fakemotor import Fake_motor
from .motorkit_pump import Motorkit_pump
from .oriental import Oriental
from .servostar import Servostar
from .pololu_tic import Pololu_tic
from .tra6ppd import Tra6ppd

actuator_list = MetaActuator.classes



# coding: utf-8

import time
from time import sleep, time
from re import fullmatch, findall
from typing import Union, Optional, List
from .inout import InOut
from .._global import OptionalModule

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  import spidev
except (ModuleNotFoundError, ImportError):
  spidev = OptionalModule("spidev")

# ADS1256 gain channel
Ads1256_gain = {1: 0b000,
                2: 0b001,
                4: 0b010,
                8: 0b011,
                16: 0b100,
                32: 0b101,
                64: 0b110}

# ADS1256 data rate
Ads1256_drate = {2.5: 0x03,
                 5: 0x13,
                 10: 0x20,
                 15: 0x33,
                 25: 0x43,
                 30: 0x53,
                 50: 0x63,
                 60: 0x72,
                 100: 0x82,
                 500: 0x92,
                 1000: 0xA1,
                 2000: 0xB0,
                 3750: 0xC0,
                 7500: 0xD0,
                 15000: 0xE0,
                 30000: 0xF0}

# Ads1256 registration definition
Ads1256_reg = {'REG_STATUS': 0,
               'REG_MUX': 1,
               'REG_ADCON': 2,
               'REG_DRATE': 3,
               'REG_IO': 4,
               'REG_OFC0': 5,
               'REG_OFC1': 6,
               'REG_OFC2': 7,
               'REG_FSC0': 8,
               'REG_FSC1': 9,
               'REG_FSC2': 10}

# Ads1256 command definition
Ads1256_cmd = {'CMD_WAKEUP': 0x00,
               'CMD_RDATA': 0x01,
               'CMD_RDATAC': 0x03,
               'CMD_SDATAC': 0x0F,
               'CMD_RREG': 0x10,
               'CMD_WREG': 0x50,
               'CMD_SELFCAL': 0xF0,
               'CMD_SELFOCAL': 0xF1,
               'CMD_SELFGCAL': 0xF2,
               'CMD_SYSOCAL': 0xF3,
               'CMD_SYSGCAL': 0xF4,
               'CMD_SYNC': 0xFC,
               'CMD_STANDBY': 0xFD,
               'CMD_RESET': 0xFE}

# Dac8532 channels definition
Dac8532_chan = {0: 0x10,
                1: 0x24}

# Waveshare AD/DA pins definition
AD_DA_pins = {'RST_PIN_ADS': 18,
              'CS_PIN_ADS': 22,
              'DRDY_PIN_ADS': 17,
              'CS_PIN_DAC': 23}


class Waveshare_ad_da(InOut):
  """Class for controlling Waveshare's AD/DA hat.

  The Waveshare_ad_da InOut block is meant for communicating with Waveshare's
  AD/DA Raspberry Pi hat, using the SPI protocol and the GPIOs. It allows to
  read values from the 8-channels ADC and/or to set the 2-channels DAC.

  Warning:
    This class is specifically meant to be used on a Raspberry Pi. See
    :ref:`Waveshare AD/DA FT232H` for use with FTDI's FT232H.
  """

  def __init__(self,
               dac_channels: Optional[List[str]] = None,
               adc_channels: Optional[List[str]] = None,
               gain_hardware: int = 1,
               v_ref: float = 3.3,
               gain: float = 1,
               offset: float = 0,
               sample_rate: Union[int, float] = 100) -> None:
    """Checks the validity of the arguments.

    Args:
      dac_channels (:obj:`list`, optional): A :obj:`list` of :obj:`str`
        representing the channels to be set. The syntax for each string is
        'DACi' with i being either `0` or `1`.
      adc_channels (:obj:`list`, optional): A :obj:`list` of :obj:`str`
        representing the channels to read. The syntax for all strings is
        either:
        ::

          'ADi' (i in range(8))

        or else:
        ::

          'ADi - ADj' (i, j in range(8))

      gain_hardware (:obj:`int`, optional): The gain to be used by the
        programmable gain amplifier. Setting a high gain allows to read small
        voltages with a better precision, but it might saturate the sensor for
        higher voltages. The available gain values are:
        ::

          1, 2, 4, 8, 16, 32, 64

      v_ref (:obj:`float`, optional): The voltage reference set by the `VREF`
        jumper. When reading single inputs, ``v_ref`` is the value the ADC
        compares the signals with. In a similar way, the maximum output voltage
        of the DAC is ``v_ref``. `3.3` and `5` are the only possible values for
        this setting, as the Raspberry Pi can only provide `3.3V` and `5V`.
      gain (:obj:`float`, optional): Allows to tune the output values of the
        DAC according to the formula:
        ::

          output = gain * tension + offset.

        The same gain applies to all the outputs.
      offset (:obj:`float`, optional): Allows to tune the output values of the
        ADC according to the formula:
        ::

          output = gain * tension + offset.

        The same offset applies to all the outputs.
      sample_rate (optional): The ADC data output rate in SPS. The available
        values are:
        ::

          2.5, 5, 10, 15, 25, 30, 50, 60, 100, 500,
          1000, 2000, 3750, 7500, 15000, 30000

    Warning:
      - ``adc_channels``:
        For reading single inputs the `JMP_AGND` jumper should normally be
        connected, whereas it should be disconnected for reading differential
        inputs. It is however possible to set a different reference than `AGND`
        for single input measurements, in which case the `JMP_AGND` jumper
        should not be connected and the voltage reference should be plugged
        in the `AINCOM` pin.

        The AD/DA offers the possibility to read single inputs or
        differential inputs, but not both at the same time ! This is due to
        the `JMP_AGND` jumper.
        For measuring both input types simultaneously, is it necessary to
        connect `AGND` to one of the channels (for example `AD0`). Then all
        single inputs `'ADi'` should be replaced by `'ADi - AD0'`. They are
        then considered as differential inputs.

        The ADC channels voltages should not be lower than `AGND-0.1V`, and not
        be greater than `AGND+5.1V`. This is independent of `VREF` value.

    Note:
      - ``adc_channels``:
        If multiple channels to read are given, they are read in a sequential
        way. This means that there's a small delay between each acquisition,
        and the timeframe is thus less accurate for the last channels than
        for the first ones. If time precision matters it is preferable to
        read as few channels as possible !

      - ``vref``:
        `VREF` can be set independently of the chosen `VCC` value. The `VCC`
        value has no influence on the ADC behaviour as it is always powered
        up with `5V`. Same goes for the DAC.
    """

    InOut.__init__(self)

    if gain_hardware not in Ads1256_gain:
      raise ValueError("gain_hardware should be in {}".format(list(
        Ads1256_gain.keys())))
    else:
      self._gain_hardware = gain_hardware

    if sample_rate not in Ads1256_drate:
      raise ValueError("sample_rate should be in {}".format(list(
        Ads1256_drate.keys())))
    else:
      self._sample_rate = sample_rate

    if v_ref not in [3.3, 5]:
      raise ValueError("v_ref should be either 3.3 or 5")
    else:
      self._v_ref = v_ref

    self._channel_set = False

    if dac_channels is None and adc_channels is None:
      print("Warning ! The AD/DA doesn't read nor write anything.")

    self._channels_write = []
    if dac_channels is not None:
      for chan in dac_channels:
        # Checking if the format matches
        if fullmatch(r'DAC[0-1]', chan) is None:
          raise ValueError("Valid format for dac_channels values is 'DACi' "
                           "with i either 0 or 1")
        else:
          self._channels_write.append(int(findall(r'\d', chan)[0]))

    self._channels_read = []
    if adc_channels is not None:
      for chan in adc_channels:
        # Checking if the format matches
        if fullmatch(r'AD[0-7]', chan) is None and \
              fullmatch(r'AD[0-7]\s?-\s?AD[0-7]', chan) is None:
          raise ValueError("Valid formats for adc_channels values are "
                           "either 'ADi' (i in range(8)) or 'ADi - ADj'")
        else:
          # Extracting the channel numbers
          chan = [int(i) for i in findall(r'\d', chan)]
          # If only one channel provided, it has to be compared to GND
          if len(chan) == 1:
            chan.append(8)
          # Making sure the two channels are not identical
          if chan[0] == chan[1]:
            raise ValueError("The two channels are the same !")
          self._channels_read.append(chan)

    self._gain = gain
    self._offset = offset

    self._bus = spidev.SpiDev(0, 0)

  def open(self) -> None:
    """Sets the SPI communication, the GPIOs and the device."""

    # Setting the GPIOs for communicating with the AD/DA
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(AD_DA_pins['RST_PIN_ADS'], GPIO.OUT)
    GPIO.setup(AD_DA_pins['CS_PIN_ADS'], GPIO.OUT)
    GPIO.setup(AD_DA_pins['DRDY_PIN_ADS'], GPIO.IN,
               pull_up_down=GPIO.PUD_UP)
    GPIO.setup(AD_DA_pins['CS_PIN_DAC'], GPIO.OUT)

    # Setting the SPI
    self._bus.max_speed_hz = 40000
    self._bus.mode = 1
    self._bus.no_cs = True
    self._reset()

    # Setting the ADS according to the user parameters
    buf = [Ads1256_gain[self._gain_hardware], Ads1256_drate[self._sample_rate]]
    GPIO.output(AD_DA_pins['CS_PIN_ADS'], GPIO.LOW)
    self._bus.writebytes([Ads1256_cmd['CMD_WREG'] |
                          Ads1256_reg['REG_ADCON'], 0x01] + buf)
    GPIO.output(AD_DA_pins['CS_PIN_ADS'], GPIO.HIGH)
    sleep(0.001)

  def get_data(self) -> List[float]:
    """Reads data from all the user-specified ADC channels, in a sequential
    way.

    Data is returned in Volts, but this can be tuned using gain and offset.

    Returns:
      :obj:`list`: A list containing the timeframe, and then the values for
      each channel to read
    """

    out = [time()]

    # The values are read one channel after the other, not simultaneously
    for chan in self._channels_read:
      GPIO.output(AD_DA_pins['CS_PIN_ADS'], GPIO.LOW)

      # Switching channel only if necessary, except for the first loop
      if len(self._channels_read) > 1 or not self._channel_set:
        self._bus.writebytes([Ads1256_cmd['CMD_WREG'] |
                              Ads1256_reg['REG_MUX'], 0x00,
                              (chan[0] << 4) | chan[1]])

        # The ADS has to be synchronized again when switching channel
        self._bus.writebytes([Ads1256_cmd['CMD_SYNC']])
        self._bus.writebytes([Ads1256_cmd['CMD_WAKEUP']])

        self._channel_set = True

      # Reading the output value
      self._wait_drdy()
      self._bus.writebytes([Ads1256_cmd['CMD_RDATA']])
      buf = self._bus.readbytes(3)
      GPIO.output(AD_DA_pins['CS_PIN_ADS'], GPIO.HIGH)

      # Converting the raw output into Volts
      out_raw = (buf[0] << 16) | (buf[1] << 8) | buf[2]
      if out_raw & 0x800000:
        out_raw -= 2 ** 24
      value = self._v_ref / self._gain_hardware * out_raw / 2 ** 23
      out.append(self._offset + self._gain * value)

    return out

  def set_cmd(self, *cmd: float) -> None:
    """Sets the user-specified DAC channels according to the input values.

    Args:
      cmd (:obj:`float`): The input values, in Volts
    """

    # The values are set one channel after the other, not simultaneously
    for val, channel in zip(cmd, self._channels_write):
      if not 0 <= val <= self._v_ref:
        raise ValueError("Desired output voltage should be between 0 and "
                         "v_ref")
      digit = int((2 ** 16 - 1) * val / self._v_ref)
      GPIO.output(AD_DA_pins['CS_PIN_DAC'], GPIO.LOW)
      self._bus.writebytes([Dac8532_chan[channel], digit >> 8,
                            digit & 0xFF])
      GPIO.output(AD_DA_pins['CS_PIN_DAC'], GPIO.HIGH)

  def close(self) -> None:
    """Releases the GPIOs."""

    self._bus.close()
    GPIO.cleanup()

  @staticmethod
  def _reset() -> None:
    """Resets the ADC."""

    GPIO.output(AD_DA_pins['RST_PIN_ADS'], GPIO.HIGH)
    sleep(0.2)
    GPIO.output(AD_DA_pins['RST_PIN_ADS'], GPIO.LOW)
    sleep(0.2)
    GPIO.output(AD_DA_pins['RST_PIN_ADS'], GPIO.HIGH)

  @staticmethod
  def _wait_drdy() -> None:
    """Waits until the `DRDY` pin is driven low, meaning that an ADC conversion
    is ready."""

    t0 = time()
    while GPIO.input(AD_DA_pins['DRDY_PIN_ADS']):
      if time() - t0 > 1:
        raise TimeoutError("Couldn't get conversion result from the ADC")



# coding: utf-8

import tables
import sys
# import numpy as np
import matplotlib.pyplot as plt

NODE = "table"  # Node containing the array to read
# If the range node is not specified, the digital levels (ints) will
# be returned
# If specified, it will use it to turn these levels in mV (SLOWER!)
# RANGE_NODE = None # The name of the node storing the ranges of each channel
RANGE_NODE = "factor"

# Change here to ignore the prompt
filename = None
start = 0  # Start to read from
stop = None  # Where to stop (leave none to read the whole file)
step = None  # Step (if None, will be computed and suggested)

filename = filename if len(sys.argv) == 1 else sys.argv[-1]
if not filename:
  filename = input("File name?")
h = tables.open_file(filename)
print(h)

arr = getattr(h.root, NODE)
lines, rows = arr.shape
print(lines, "lines and", rows, "rows")
a = rows * lines // 500000 or 1
if a > 1:
  print("How many lines do you want to skip on each read?")
  print("1 will read ALL the data (may use too much memory!)")
  print("Suggested value:", a)
  if not step:
    step = int(raw_input("Read one out of how many lines?({})".format(a)) or a)
else:
  step = a

i = 0
if not hasattr(h.root, RANGE_NODE):
  RANGE_NODE = None
print("Reading...")
if RANGE_NODE:
  out = arr.read(start=start, stop=stop, step=step).astype(float)
else:
  out = arr.read(start=start, stop=stop, step=step)
print(out.shape)

if RANGE_NODE:
  ranges = getattr(h.root, RANGE_NODE).read()
  h.close()
  print("Applying factor...")
  for i, r in enumerate(ranges):
    out[:, i] *= r
else:
  h.close()

print("Plotting...")
for i in range(rows):
  plt.plot(out[:, i])
print("Done!")
plt.show()



# coding: utf-8

from time import sleep
from typing import Optional
from .actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")

ACCEL = b'.1'  # Acceleration and deceleration times


class Oriental(Actuator):
  """To drive an axis with an oriental motor through a serial link.

  The current setup moves at `.07mm/min` with `"VR 1"`.
  """

  def __init__(self,
               baudrate: int = 115200,
               port: str = '/dev/ttyUSB0',
               gain: float = 1/.07) -> None:
    """Sets the instance attributes.

    Args:
      baudrate (:obj:`int`, optional): Set the corresponding baud rate.
      port (:obj:`str`, optional): Path to connect to the serial port.
      gain (:obj:`float`, optional): The gain for speed commands.
    """

    Actuator.__init__(self)
    self.baudrate = baudrate
    self.port = port
    self.speed = 0
    self.gain = gain  # unit/(mm/min)

  def open(self) -> None:
    self.ser = serial.Serial(self.port, baudrate=self.baudrate, timeout=0.1)
    for i in range(1, 5):
      self.ser.write("TALK{}\n".format(i).encode('ASCII'))
      ret = self.ser.readlines()
      if "{0}>".format(i).encode('ASCII') in ret:
        self.num_device = i
        motors = ['A', 'B', 'C', 'D']
        print("Motor connected to port {} is {}".format(self.port,
                                                        motors[i-1]))
        break
    self.clear_errors()
    self.ser.write(b"TA " + ACCEL+b'\n')  # Acceleration time
    self.ser.write(b"TD " + ACCEL+b'\n')  # Deceleration time

  def clear_errors(self) -> None:
    self.ser.write(b"ALMCLR\n")

  def close(self) -> None:
    self.stop()
    self.ser.close()

  def stop(self) -> None:
    """Stop the motor."""

    self.ser.write(b"SSTOP\n")
    sleep(float(ACCEL))
    # sleep(1)
    self.speed = 0

  def reset(self) -> None:
    self.clear_errors()
    self.ser.write(b"RESET\n")
    self.ser.write("TALK{}\n".format(self.num_device).encode('ASCII'))
    self.clear_errors()

  def set_speed(self, cmd: float) -> None:
    """Pilot in speed mode, requires speed in `mm/min`."""

    # speed in mm/min
    # gain can be edited by giving gain=xx to the init
    speed = min(100, int(abs(cmd * self.gain) + .5))  # Closest value < 100
    # These motors take ints only
    if speed == 0:
      self.stop()
      return
    sign = int(self.gain * cmd / abs(self.gain * cmd))
    signed_speed = sign * speed
    if signed_speed == self.speed:
      return
    dirchg = self.speed * sign < 0
    if dirchg:
      # print("DEBUGORIENTAL changing dir")
      self.stop()
    self.ser.write("VR {}\n".format(abs(speed)).encode('ASCII'))
    if sign > 0:
      # print("DEBUGORIENTAL going +")
      self.ser.write(b"MCP\n")
    else:
      # print("DEBUGORIENTAL going -")
      self.ser.write(b"MCN\n")
    self.speed = signed_speed

  def set_home(self) -> None:
    self.ser.write(b'preset\n')

  def move_home(self) -> None:
    self.ser.write(b'EHOME\n')

  def set_position(self,
                   position: float,
                   speed: Optional[float] = None) -> None:
    """Pilot in position mode, needs speed and final position to run
    (in `mm/min` and `mm`)."""

    if speed is None:
      raise ValueError("The Oriental actuator needs both a position and a "
                       "speed command when driven in position mode !")

    self.ser.write("VR {0}".format(abs(speed)).encode('ASCII'))
    self.ser.write("MA {0}".format(position).encode('ASCII'))

  def get_position(self) -> float:
    """Reads current position.

    Returns:
      Current position of the motor.
    """

    self.ser.flushInput()
    self.ser.write(b'PC\n')
    self.ser.readline()
    actuator_pos = self.ser.readline()
    actuator_pos = str(actuator_pos)
    try:
      actuator_pos = float(actuator_pos[4:-3])
    except ValueError:
      print("PositionReadingError")
      return 0
    return actuator_pos



# coding: utf-8

from time import time
from typing import List
from .inout import InOut
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Opsens(InOut):
  """This class allows reading data from an Opsens PicoSens fiber optics signal
  conditioner.

  It can read data from various fiber optics sensors like temperature,
  pressure, position or strain.
  """

  def __init__(self,
               device: str = '/dev/ttyUSB0') -> None:
    """Sets the arg and initializes the parent class.

    Args:
      device: Address of the serial connection for communicating with the
        PicoSens.
    """

    super().__init__()

    self._addr = device
    self._dev = None

  def open(self) -> None:
    """Opens the serial connection and configures the PicoSens."""

    self._dev = serial.Serial(port=self._addr, baudrate=57600)
    self._send_cmd("meas:rate min")

  def get_data(self) -> List[float]:
    """Reads data from the PicoSens and returns it."""

    return [time(), float(self._send_cmd("ch1:data? 1")[:-3])]

  def close(self) -> None:
    """Closes the serial connection if it was opened."""

    if self._dev is not None:
      self._dev.close()

  def _send_cmd(self, cmd: str) -> str:
    """Sends a command and returns the received answer."""

    self._dev.write(cmd + '\n')
    return self._dev.read_until(b'\x04\n').decode()



# coding: utf-8

import numpy as np
from typing import Dict, Any
from .modifier import Modifier


class Moving_med(Modifier):
  """Modifier replacing the data of each label with its median value over a
  chosen number of points.

  Unlike :ref:`Median`, it returns a value each time a message is received from
  the upstream link.
  """

  def __init__(self, n_points: int = 100) -> None:
    """Sets the args and initializes the parent class.

    Args:
      n_points: The maximum number of points on which to compute the median.
    """

    super().__init__()
    self._n_points = n_points
    self._buf = None

  def evaluate(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """Receives data from the upstream block, computes the median of every
    label and replaces the original data with it."""

    # Initializing the buffer
    if self._buf is None:
      self._buf = {key: [value] for key, value in data.items()}

    ret = {}
    for label in data:
      # Updating the buffer with the newest data
      self._buf[label].append(data[label])

      # Trimming the buffer if there's too much data
      if len(self._buf[label]) > self._n_points:
        self._buf[label] = self._buf[label][-self._n_points:]

      # Calculating the median for each label
      try:
        ret[label] = np.median(self._buf[label])
      except TypeError:
        ret[label] = self._buf[label][-1]

    return ret
